import logging

import numpy as np
from qtpy.QtCore import Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtCore import QEvent, Qt
from qtpy.QtGui import QColorConstants, QIcon, QKeyEvent
from qtpy.QtWidgets import (
    QComboBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QRadioButton,
    QToolButton,
    QVBoxLayout,
)
from scipy.interpolate import BSpline, make_interp_spline

from idtrackerai import ListOfBlobs
from idtrackerai.GUI_tools import (
    CanvasMouseEvent,
    CanvasPainter,
    LightPopUp,
    WrappedLabel,
    get_icon,
    key_event_modifier,
)


class CustomComboBox(QComboBox):
    def keyPressEvent(self, e: QKeyEvent | None) -> None:
        if e is None:
            return
        event = key_event_modifier(e)
        if event is not None:
            super().keyPressEvent(event)

    def keyReleaseEvent(self, e: QKeyEvent | None) -> None:
        if e is None:
            return
        event = key_event_modifier(e)
        if event is not None:
            super().keyReleaseEvent(event)


class Interpolator(QGroupBox):
    interpolation_kinds = {"Linear": 1, "Quadratic": 2, "Cubic": 3, "5th order": 5}
    neew_to_draw = Signal()
    update_trajectories = Signal(int, int, bool)  # start, end, update_errors
    go_to_frame = Signal(int)
    preload_frames = Signal(int, int)
    interpolation_accepted = Signal()
    enabled_changed = Signal(bool)
    interp_spline: BSpline
    interp_frames: np.ndarray
    interp_points: np.ndarray

    def __init__(self) -> None:
        self.popup = LightPopUp()
        super().__init__()
        layout = QVBoxLayout()
        self.setLayout(layout)

        self.warning = WrappedLabel()
        self.warning.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(self.warning)
        self.warning.setVisible(False)

        self.goto_btn = QPushButton()
        layout.addWidget(self.goto_btn)
        self.goto_btn.setVisible(False)

        self.info_label = WrappedLabel()
        layout.addWidget(self.info_label)

        range_row = QHBoxLayout()
        range_row.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.start_btn = QToolButton()
        self.end_btn = QToolButton()
        self.start_btn.clicked.connect(lambda: self.go_to_frame.emit(self.start - 1))
        self.end_btn.clicked.connect(lambda: self.go_to_frame.emit(self.end))
        range_row.addWidget(QLabel("From"))
        range_row.addWidget(self.start_btn)
        range_row.addWidget(QLabel("to"))
        range_row.addWidget(self.end_btn)
        layout.addLayout(range_row)

        self.interpolation_order_box = CustomComboBox()
        self.interpolation_order_box.addItems(self.interpolation_kinds.keys())
        self.interpolation_order_box.setCurrentText("Cubic")
        self.interpolation_order_box.currentTextChanged.connect(self.new_interp_type)
        order_row = QHBoxLayout()
        self.interpolation_order_label = WrappedLabel("Spline order")
        order_row.addWidget(self.interpolation_order_label)
        order_row.addWidget(self.interpolation_order_box)
        layout.addLayout(order_row)

        self.input_size_row = QHBoxLayout()
        self.input_size_row.addWidget(WrappedLabel("Real values"))
        for value in (10, 100, 1000):
            btn = QRadioButton(str(value))
            if value == 10:
                btn.setChecked(True)
            btn.clicked.connect(self.new_input_size)
            self.input_size_row.addWidget(btn)
        layout.addLayout(self.input_size_row)
        self.input_size = 10

        remove_centroid = QPushButton("Remove centroid [R]")
        remove_centroid.setIcon(QIcon.fromTheme("user-trash"))
        remove_centroid.setShortcut(Qt.Key.Key_R)
        remove_centroid.clicked.connect(self.remove_current_centroid)
        layout.addWidget(remove_centroid)

        apply_row = QHBoxLayout()

        self.abort_btn = QPushButton(get_icon("cancel"), "Abort [Esc]")
        self.abort_btn.setShortcut(Qt.Key.Key_Escape)
        self.abort_btn.clicked.connect(self.abort_interpolation)
        apply_row.addWidget(self.abort_btn)

        self.apply_btn = QPushButton(get_icon("ok"), "Apply [Ctrl+A]")
        self.apply_btn.setShortcut("Ctrl+A")
        self.apply_btn.clicked.connect(self.apply_interpolation)
        apply_row.addWidget(self.apply_btn)

        layout.addLayout(apply_row)

        self.setActivated(False)
        self.animal_id: int = -1

    def trajectories_have_been_updated(self) -> None:
        if self.isEnabled():
            self.expand_start()
            self.expand_end()
            self.build_interpolator()

    def changeEvent(self, a0: QEvent | None) -> None:
        if a0 is None:
            return
        if a0.type() == QEvent.Type.EnabledChange:
            self.enabled_changed.emit(self.isEnabled())

    def new_interp_type(self, kind: str) -> None:
        self.interp_spline = make_interp_spline(
            self.interp_frames,
            self.interp_points,
            k=self.interpolation_kinds[kind],
            check_finite=False,
        )
        self.neew_to_draw.emit()

    def new_input_size(self) -> None:
        btn = self.sender()
        assert isinstance(btn, QRadioButton)
        self.input_size = int(btn.text())
        self.build_interpolator()

    def set_interpolation_params(self, animal_id: int, start: int, end: int) -> None:
        self.start = start
        self.end = end
        self.animal_id = animal_id - 1
        self.expand_start()
        self.expand_end()
        self.build_interpolator()
        self.preload_frames.emit(
            # this makes the video player to preload some frames that the user might want to see
            max(0, self.start - 5),
            min(self.end + 10, self.start + 20, self.n_frames),
        )

    def build_interpolator(self) -> None:
        self.interpolation_range = range(self.start, self.end)
        self.continuous_interpolation_range = np.arange(
            max(self.start - 1, 0), self.end + 0.1, 0.2
        )
        self.entire_range = range(
            max(0, self.start - self.input_size),
            min(self.n_frames, self.end + self.input_size),
        )

        # if there are not enough non-NaN values in self.entire_range,
        # the interpolator cannot be initialized so we expend the range
        finding_non_nans_iterations = 0
        while (
            np.isfinite(self.trajectories[self.entire_range, self.animal_id, 0]).sum()
            < 10
        ):
            self.entire_range = range(
                max(0, self.entire_range.start - 10),
                min(self.n_frames, self.entire_range.stop + 10),
            )

            finding_non_nans_iterations += 1
            if finding_non_nans_iterations > 10:
                break

        n_duplicated = np.count_nonzero(
            self.duplicated[self.entire_range, self.animal_id]
        )
        if n_duplicated:
            first_duplicated = (
                self.duplicated[self.entire_range, self.animal_id].argmax()
                + self.entire_range.start
            )
            self.warning.setText(
                f'<font color="red">There are {n_duplicated} frames where identity'
                f" {self.animal_id + 1} appears duplicated. It is highly recommended to"
                " solve this before proceeding. The first duplication appears at frame"
                f" {first_duplicated}"
            )
            self.goto_btn.setText(f"Go to {first_duplicated}")
            self.goto_btn.clicked.connect(
                lambda: self.go_to_frame.emit(first_duplicated)
            )
            self.warning.setVisible(True)
            self.goto_btn.setVisible(True)
        else:
            self.warning.setVisible(False)
            self.goto_btn.setVisible(False)

        self.interp_frames = np.asarray(self.entire_range)[
            np.isfinite(self.trajectories[self.entire_range, self.animal_id, 0])
        ]
        try:
            self.interp_points = self.trajectories[self.interp_frames, self.animal_id]
            self.interp_spline = make_interp_spline(
                self.interp_frames,
                self.interp_points,
                k=self.interpolation_kinds[self.interpolation_order_box.currentText()],
                check_finite=False,
            )
        except ValueError as exc:
            self.setActivated(False)
            logging.error("Unexpected error", exc_info=exc)
            QMessageBox.warning(
                self,
                "Unexpected error",
                f"Unexpected error while initializing interpolation:\n{exc}",
            )
        else:
            self.setActivated(True)
            self.info_label.setText(
                "Interpolating identity <span"
                f' style="font-weight:600">{self.animal_id + 1}'
            )

    def remove_current_centroid(self) -> None:
        if self.current_frame not in self.entire_range:
            QMessageBox.warning(
                self,
                "Interpolator message",
                "Cannot remove current centroid outside interpolation "
                f"range ({self.entire_range.start} -> {self.entire_range.stop})",
            )
            return

        centroid_to_remove = self.trajectories[self.current_frame, self.animal_id]
        if np.isnan(centroid_to_remove[0]):
            self.popup.warning(
                "Interpolator message",
                "Cannot remove current centroid because it does not exist",
            )
            return

        self.list_of_blobs.remove_centroid(
            self.current_frame, centroid_to_remove, self.animal_id + 1
        )

        # This signal is connected to ValidationGUI.update_trajectories_range()
        # which calls self.trajectories_have_been_updated() to re-build the interpolator.
        # This is why we don't call self.build_interpolator() in here.
        self.update_trajectories.emit(self.current_frame, self.current_frame + 1, False)

    def expand_start(self) -> None:
        for frame in range(self.start - 1, -1, -1):
            if (
                not np.isnan(self.trajectories[frame, self.animal_id, 0])
                or not self.in_tracking_interval[frame]
            ):
                if frame + 1 != self.start:
                    self.start = frame + 1
                    self.go_to_frame.emit(frame)
                return

        # we haven't found any non nan value, interpolation starts at the beginning of the video
        if self.start != 0:  # we emit go_to_frame only if start value changed
            self.start = 0
            self.go_to_frame.emit(0)

    def expand_end(self) -> None:
        for frame in range(self.end, self.n_frames):
            if (
                not np.isnan(self.trajectories[frame, self.animal_id, 0])
                or not self.in_tracking_interval[frame]
            ):
                if frame != self.end:
                    self.end = frame
                    self.go_to_frame.emit(frame)
                return

        # we haven't found any non nan value, interpolation ends at the end of the video
        if self.end != self.n_frames:  # we emit go_to_frame only if end value changed
            self.end = self.n_frames
            self.go_to_frame.emit(self.n_frames)

    def click_event(self, event: CanvasMouseEvent) -> None:
        if (
            event.button != Qt.MouseButton.RightButton
            or not self.isEnabled()
            or self.current_frame not in self.interpolation_range
        ):
            return

        current_postion = self.trajectories[self.current_frame, self.animal_id]
        already_has_a_centroid = not np.isnan(current_postion[0])
        if already_has_a_centroid:
            # TODO do not use self.trajectories but always try to update the
            # centroid or add a new one if no centroid is found
            self.list_of_blobs.update_centroid(
                self.current_frame, self.animal_id + 1, current_postion, event.xy_data
            )
        else:
            self.list_of_blobs.add_centroid(
                self.current_frame, self.animal_id + 1, event.xy_data
            )
        self.update_trajectories.emit(self.current_frame, self.current_frame + 1, False)

    def setActivated(self, activated: bool) -> None:
        self.setEnabled(activated)
        if not activated:
            self.warning.setVisible(False)
            self.goto_btn.setVisible(False)
            self.info_label.setText(
                'Select some errors of kind "Miss id" of '
                '"Jump" to start an interpolation process'
            )
        self.neew_to_draw.emit()

    def abort_interpolation(self) -> None:
        logging.debug("Abort interpolation")
        self.update_trajectories.emit(self.start, self.end, True)
        self.setActivated(False)

    def apply_interpolation(self) -> None:
        logging.debug("Apply interpolation")
        for new_centroid, frame in zip(
            self.interp_spline(self.interpolation_range), self.interpolation_range
        ):
            if np.isnan(self.trajectories[frame, self.animal_id, 0]):
                self.list_of_blobs.add_centroid(frame, self.animal_id + 1, new_centroid)
        self.setActivated(False)
        self.interpolation_accepted.emit()
        self.update_trajectories.emit(self.start, self.end, True)

    @property
    def start(self) -> int:
        return self._start

    @start.setter
    def start(self, value: int) -> None:
        self._start = value
        self.start_btn.setText(f"frame {value - 1}")
        self.start_btn.setToolTip(f"Go to frame {value - 1}")

    @property
    def end(self) -> int:
        return self._end

    @end.setter
    def end(self, value: int) -> None:
        self._end = value
        self.end_btn.setText(f"frame {value}")
        self.end_btn.setToolTip(f"Go to frame {value}")

    def set_references(
        self,
        traj: np.ndarray,
        unidentified: np.ndarray,
        duplicated: np.ndarray,
        list_of_blobs: ListOfBlobs,
        tracking_intervals: list[list[int]] | None,
    ) -> None:
        self.list_of_blobs = list_of_blobs
        self.trajectories = traj
        self.unidentified = unidentified
        self.duplicated = duplicated
        self.n_frames = len(self.trajectories)
        if tracking_intervals is None:
            self.in_tracking_interval = np.ones(len(traj), bool)
        else:
            self.in_tracking_interval = np.zeros(len(traj), bool)
            for start, end in tracking_intervals:
                self.in_tracking_interval[start:end] = True
        self.setActivated(False)

    def paint_on_canvas(self, painter: CanvasPainter, frame: int) -> None:
        self.current_frame = frame

        # interpolated points
        painter.setPen(QColorConstants.White)
        painter.setBrush(QColorConstants.White)
        for point in self.interp_spline(self.interpolation_range):
            painter.drawBigPoint(*point)

        # continuum interpolated range
        painter.drawPolylineFromVertices(
            self.interp_spline(self.continuous_interpolation_range)
        )

        # interpolator input data
        painter.setPen(QColorConstants.Red)
        painter.setBrush(QColorConstants.Red)
        painter.drawPolylineFromVertices(
            self.interp_points[self.interp_frames < self.start]
        )
        painter.drawPolylineFromVertices(
            self.interp_points[self.interp_frames >= self.end]
        )
        for point in self.interp_points:
            painter.drawBigPoint(*point)

        # actual point
        if (
            self.current_frame in self.interp_frames
            or self.current_frame in self.interpolation_range
        ):
            painter.setPen(QColorConstants.White)
            painter.drawBigPoint(*self.interp_spline(frame))



import gc
import logging
from itertools import pairwise
from pprint import pformat

import numpy as np

from idtrackerai import ListOfBlobs, ListOfFragments, ListOfGlobalFragments, Session
from idtrackerai.base.network import IdentifierBase
from idtrackerai.utils import track


def tracker_API(
    session: Session,
    list_of_blobs: ListOfBlobs,
    list_of_fragments: ListOfFragments,
    list_of_global_fragments: ListOfGlobalFragments,
) -> IdentifierBase | None:

    if session.track_wo_identities:
        track_without_identities(session, list_of_blobs, list_of_fragments)
        return None

    if session.single_animal:
        track_single_animal(list_of_blobs)
        return None

    if len(list_of_fragments) == 1:
        logging.warning("Tracking a single fragment")
        for blob in list_of_blobs.all_blobs:
            if blob.fragment_identifier == list_of_fragments.fragments[0].identifier:
                blob.identity = 1
        return None

    from .tracker import run_tracker

    logging.info(
        "Deleting ListOfBlobs to save memory, it will be reloaded from disk after"
        " tracking"
    )
    for blob in list_of_blobs.all_blobs:
        blob.__dict__.clear()
    list_of_blobs.blobs_in_video.clear()
    # Blobs contain circular references between them, so the automatic garbage
    # collector won't delete them immediately after the clear().
    # Manually calling gc.collect() is the way to really free RAM
    gc.collect()
    identifier_model = run_tracker(session, list_of_fragments, list_of_global_fragments)
    list_of_fragments.update_id_images_dataset()
    gc.collect()  # just in case
    list_of_blobs.blobs_in_video = ListOfBlobs.load(session.blobs_path).blobs_in_video
    return identifier_model


def track_single_animal(list_of_blobs: ListOfBlobs):
    logging.info("Tracking a single animal, assigning identity 1 to all blobs")
    for blob in list_of_blobs.all_blobs:
        blob.identity = 1


def track_without_identities(
    session: Session, list_of_blobs: ListOfBlobs, list_of_fragments: ListOfFragments
):
    logging.info("Tracking without identities")
    roi_to_ids: dict[int, set[int]] = {}
    if session.exclusive_rois:  # Assign identities based on exclusive ROIs if present
        all_rois = {frag.exclusive_roi for frag in list_of_fragments}
        n_blobs_per_region = np.zeros(
            (list_of_blobs.number_of_frames, max(all_rois) + 1), dtype=int
        )

        for fragment in list_of_fragments.individual_fragments:
            n_blobs_per_region[
                fragment.start_frame : fragment.end_frame, fragment.exclusive_roi
            ] += 1

        max_blobs_per_region = n_blobs_per_region.max(axis=0)
        counter = 0
        for roi, max_blobs in enumerate(max_blobs_per_region):
            roi_to_ids[roi] = set(range(counter, counter + max_blobs))
            counter += max_blobs

        session.number_of_animals = max_blobs_per_region.sum()
        session.identities_groups = {
            f"Region_{roi}": {id_ + 1 for id_ in ids} for roi, ids in roi_to_ids.items()
        }
        logging.info(
            "Identity groups by exclusive ROIs:\n%s", pformat(session.identities_groups)
        )
    else:
        session.number_of_animals = max(
            len(frame) for frame in list_of_blobs.blobs_in_video
        )
        roi_to_ids[-1] = set(range(session.number_of_animals))

    # Initialize current fragments for each animal slot
    current_fragments = [-10 for _ in range(session.number_of_animals)]

    # Iterate through frames in pairs to assign identities
    for blobs_in_frame, blobs_in_next in pairwise(
        track(list_of_blobs.blobs_in_video + [[]], "Assigning random identities")
    ):
        next_fragments = {b.fragment_identifier for b in blobs_in_next}

        for blob in blobs_in_frame:
            if blob.is_a_crossing:
                continue

            try:
                identity = current_fragments.index(blob.fragment_identifier)
            except ValueError:
                # Find a valid identity slot for this fragment
                roi = (
                    list_of_fragments.fragments[blob.fragment_identifier].exclusive_roi
                    if session.exclusive_rois
                    else -1
                )
                valid_ids = roi_to_ids[roi]
                possible_ids = [
                    idx
                    for idx, frag_id in enumerate(current_fragments)
                    if frag_id == -10 and idx in valid_ids
                ]
                if not possible_ids:
                    logging.error(
                        f"No valid identity found for fragment {blob.fragment_identifier} in ROI {roi}"
                    )
                    continue
                identity = possible_ids[0]
                current_fragments[identity] = blob.fragment_identifier

            blob.identity = identity + 1
            if blob.fragment_identifier not in next_fragments:
                current_fragments[identity] = -10  # Free the slot



# Each Qt binding is different, so...
# pyright: reportIncompatibleMethodOverride=false
import logging
from collections.abc import Sequence

from qtpy.QtCore import Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtCore import Qt
from qtpy.QtGui import QWheelEvent
from qtpy.QtWidgets import QHBoxLayout, QSizePolicy, QSlider, QSpinBox, QWidget
from superqt import QLabeledRangeSlider


class LabelRangeSlider(QLabeledRangeSlider):
    single_value_changed = Signal(int)

    def __init__(
        self,
        min: int,
        max: int,
        parent: QWidget | None = None,
        start_end_val: tuple[int, int] | None = None,
        block_upper=True,
    ):
        self.parent_widget = parent
        super().__init__(Qt.Orientation.Horizontal, parent)
        self.block_upper = block_upper
        self.setRange(min, max)
        self.setValue(start_end_val or (min, max))
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)

        self._min_label.setReadOnly(True)
        if block_upper:
            self._max_label.setReadOnly(True)
        else:
            self._max_label.setFocusPolicy(Qt.FocusPolicy.WheelFocus)

        def wheelEvent_on_min(event: QWheelEvent) -> None:
            steps = event.angleDelta().y() // 120
            val0, val1 = self._slider.value()
            self._slider.setSliderPosition((val0 + steps, val1))

        def wheelEvent_on_max(event: QWheelEvent) -> None:
            steps = event.angleDelta().y() // 120
            val0, val1 = self._slider.value()
            self._slider.setSliderPosition((val0, val1 + steps))

        self._handle_labels[0].wheelEvent = wheelEvent_on_min
        self._handle_labels[1].wheelEvent = wheelEvent_on_max

        def send_single_value_changed(value: int) -> None:
            """Emit the single value changed signal for the first handle"""
            try:
                self.single_value_changed.emit(int(value))
            except Exception as e:
                logging.error(f"Error emitting single_value_changed signal: {e}")

        for handle in self._handle_labels:
            handle.setFocusPolicy(Qt.FocusPolicy.WheelFocus)
            handle.textChanged.connect(send_single_value_changed)

    def value(self) -> tuple[int, int]:
        return super().value()  # type: ignore

    def setValue(self, value: Sequence[int]) -> None:
        if not self.block_upper and value[1] > self.maximum():
            self.setMaximum(value[1])
        return super().setValue(value)  # type: ignore


class InvertibleSlider(QWidget):
    "A labeled slider with the capacity to invert the colored bar without negative numbers"

    valueChanged = Signal(int)

    def __init__(self, min: int, max: int) -> None:
        super().__init__()
        main_layout = QHBoxLayout()
        self.min = min
        self.max = max

        self.slider = QSlider()
        # We manage wheel events in this class
        self.slider.wheelEvent = lambda e: e.ignore() if e is not None else None
        self.slider.setOrientation(Qt.Orientation.Horizontal)
        self.slider.setRange(min, max)

        self.label = QSpinBox()
        self.label.setFocusPolicy(Qt.FocusPolicy.WheelFocus)
        self.label.setButtonSymbols(QSpinBox.ButtonSymbols.NoButtons)
        self.label.setStyleSheet("background:transparent; border: 0;")
        self.label.setRange(min, max)

        main_layout.addWidget(self.slider)
        main_layout.addWidget(self.label)
        main_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(main_layout)

        self.inverted = False

        self.slider.valueChanged.connect(self.slider_changed)
        self.label.valueChanged.connect(self.label_changed)

    def slider_changed(self, value: int):
        self.label.setValue(-value if self.inverted else value)

    def label_changed(self, value: int):
        self.slider.setValue(-value if self.inverted else value)
        self.valueChanged.emit(value)

    def value(self):
        return self.label.value()

    def setValue(self, val: int):
        return self.label.setValue(val)

    def set_inverted(self, inverted: bool):
        """This function must be followed by a valueChanged signal
        I don't implement it here to avoid duplicates and simplify code downstream"""
        self.slider.setInvertedAppearance(inverted)
        self.slider.setInvertedControls(inverted)
        current_value = self.slider.value()
        self.inverted = inverted

        self.slider.blockSignals(True)
        if inverted:
            self.slider.setRange(-self.max, self.min)
            self.slider.setValue(-abs(current_value))
        else:
            self.slider.setRange(self.min, self.max)
            self.slider.setValue(abs(current_value))
        self.slider.blockSignals(False)

    def set_value(self, value: int):
        self.label.setValue(value)

    def wheelEvent(self, e: QWheelEvent) -> None:
        steps = e.angleDelta().y()
        if steps > 0:
            self.setValue(self.value() + 1)
        elif steps < 0:
            self.setValue(self.value() - 1)
        else:
            e.ignore()
        e.accept()



import json
import logging
from abc import ABC
from collections.abc import Callable, Sequence
from pathlib import Path

import numpy as np
import torch
from torch import Tensor, nn
from torchvision.models.resnet import BasicBlock, ResNet


class ResNet18(ResNet):

    __call__: Callable[[Tensor], Tensor]

    def __init__(self, n_channels_in: int = 1, n_dimensions_out: int = 8) -> None:
        super().__init__(BasicBlock, [2, 2, 2, 2])
        if n_channels_in != 3:
            # adapt first conv layer to our single channel images (not RGB)
            self.conv1 = torch.nn.Conv2d(
                n_channels_in, 64, kernel_size=7, stride=2, padding=3, bias=False
            )
            nn.init.kaiming_normal_(
                self.conv1.weight, mode="fan_out", nonlinearity="relu"
            )

        # The last fully connected layer gives the coordinates in the embedding space
        # we do not need the bias term because only the relative distances matter
        self.fc = nn.Linear(512 * BasicBlock.expansion, n_dimensions_out, bias=False)

    @classmethod
    def from_file(cls, path: Path | str):
        logging.info(f"Loading {cls.__name__} weights from {path}")
        model_state_dict = torch.load(path, weights_only=True)
        n_dimensions_out = len(model_state_dict["fc.weight"])
        n_channels_in = model_state_dict["conv1.weight"].shape[1]
        model_state_dict.pop("fc.bias", None)  # remove bias if it exists
        model = cls(n_channels_in, n_dimensions_out)
        model.load_state_dict(model_state_dict)
        return model

    def compile(self) -> None:
        logging.info("Compiling the model %s", self.__class__.__name__)
        return super().compile()


class IdCNN(nn.Module):
    __call__: Callable[[Tensor], Tensor]
    legacy_normalization: bool = False
    """Before v6.0.0, the models were trained with a different
    normalization scheme. This flag can reenable it for backward compatibility."""

    def __init__(self, input_shape: Sequence[int], out_dim: int):
        logging.info(f"Creating {self.__class__.__name__} model")
        super().__init__()

        self.layers = nn.Sequential(
            nn.Conv2d(input_shape[-1], 16, 5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(2, stride=2),
            nn.Conv2d(16, 64, 5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(2, stride=2),
            nn.Conv2d(64, 100, 5, padding=2),
            nn.ReLU(inplace=True),
            nn.Flatten(),
            nn.Linear(100 * (input_shape[1] // 4) ** 2, 100),
            nn.ReLU(inplace=True),
            nn.Linear(100, out_dim),
        )

        self.reinitilaize()

    def forward(self, x: Tensor) -> Tensor:
        if self.legacy_normalization:
            x -= x.mean((1, 2, 3), keepdim=True)
            x /= x.std((1, 2, 3), keepdim=True)
        return self.layers(x)

    def reinitilaize(self):
        logging.info(f"Reinitializing {self.__class__.__name__}")

        def init_func(m):
            if isinstance(m, (nn.Linear, nn.Conv2d)):
                nn.init.xavier_uniform_(m.weight.data)

        self.apply(init_func)

    def fully_connected_reinitialization(self):
        logging.info(
            f"Reinitializing only {self.__class__.__name__} fully connected layers"
        )

        def init_func(m):
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight.data)

        self.apply(init_func)

    @classmethod
    def load(cls, image_size: Sequence[int], model_path: Path):
        assert model_path.is_dir()

        for name in (  # v5.0.0 compatibility
            "identifier_cnn.model.pt",
            "identification_network.model.pth",
            "identification_network_.model.pth",
            "supervised_identification_network.model.pth",
            "supervised_identification_network_.model.pth",
        ):
            if (model_path / name).is_file():
                model_path = model_path / name
                break
        else:
            raise FileNotFoundError(model_path)

        model_params_path = model_path.parent / "model_params.json"
        if model_params_path.is_file():
            model_params = json.loads(model_params_path.read_text())
            if "version" in model_params:
                version = tuple(
                    map(int, model_params["version"].split("a")[0].split("."))
                )
            else:
                # before 6.0.0 the version was not included in the model_params.json
                version = None
        else:
            if "_" in model_path.parent.name:
                # prior to 6.0.0, the accumulation folders had an index like "accumulation_0"
                version = None
            else:
                version = (6, 0, 0)  # at least 6.0.0

        logging.info("Load model weights from %s", model_path)
        # The path to model file (*.best_model.pth). Do NOT use checkpoint file here
        # model_path can contain metadata in previous versions of idtrackerai, that's why weights_only=False
        model_state: dict = torch.load(model_path, weights_only=False)
        model_state.pop("val_acc", None)
        model_state.pop("test_acc", None)
        model_state.pop("ratio_accumulated", None)

        if "fc2.weight" in model_state:
            n_classes = len(model_state["fc2.weight"])
        else:
            n_classes = len(model_state["layers.11.weight"])

        model = cls(image_size, n_classes)

        if version is None or version < (6, 0, 0):
            # The model was trained with legacy normalization, we need to set it to True
            model.legacy_normalization = True

        try:
            model.load_state_dict(model_state, strict=True)
        except RuntimeError:
            logging.warning(
                "Loading a model from a version older than 5.1.7, "
                "going to translate the state dictionary."
            )
            translated_model_state = {
                "layers.0.weight": model_state["conv1.weight"],
                "layers.0.bias": model_state["conv1.bias"],
                "layers.3.weight": model_state["conv2.weight"],
                "layers.3.bias": model_state["conv2.bias"],
                "layers.6.weight": model_state["conv3.weight"],
                "layers.6.bias": model_state["conv3.bias"],
                "layers.9.weight": model_state["fc1.weight"],
                "layers.9.bias": model_state["fc1.bias"],
                "layers.11.weight": model_state["fc2.weight"],
                "layers.11.bias": model_state["fc2.bias"],
            }
            model.load_state_dict(translated_model_state, strict=True)

        return model

    def compile(self) -> None:
        logging.info("Compiling the model %s", self.__class__.__name__)
        return super().compile()


class IdentifierBase(ABC):
    model: nn.Module
    model_weights_filename: str

    def compile(self) -> None:
        return self.model.compile()

    def __init__(self, model: nn.Module) -> None:
        self.model = model

    def eval(self) -> None:
        self.model.eval()

    def train(self) -> None:
        self.model.train()

    def to(self, device: "str | torch.device | int"):
        self.model.to(device)
        return self

    def forward(self, images: Tensor) -> Tensor:
        "Takes a tensor images of size (Batch size, 1 ,Height, Width) in the range [0,1] and outputs another tensor of size (Batch size, n_animals) for the predicted probabilities"
        raise NotImplementedError

    def __call__(self, images: Tensor) -> Tensor:
        return self.forward(images)

    @classmethod
    def load(cls, *args, **kwargs):
        raise NotImplementedError

    def save(self, path: Path) -> None:
        assert path.is_dir()
        path = path / self.model_weights_filename
        logging.info("Saving %s at %s", self.__class__.__name__, path)
        torch.save(self.model.state_dict(), path)


class IdentifierIdCNN(IdentifierBase):
    model: IdCNN  # pyright: ignore[reportIncompatibleVariableOverride] IdCNN is subclass of torch.nn.Module
    model_weights_filename: str = "identifier_cnn.model.pt"

    def forward(self, images: Tensor) -> Tensor:
        return nn.functional.softmax(self.model(images), dim=1)

    def save(self, path: Path) -> None:
        assert path.is_dir()
        return super().save(path)

    @classmethod
    def load(cls, image_size: Sequence[int], model_path: Path):
        return cls(IdCNN.load(image_size, model_path))


class IdentifierContrastive(IdentifierBase):
    cluster_centers: Tensor
    model_weights_filename: str = "identifier_contrastive.model.pt"
    cluster_centers_filename: str = "identifier_contrastive.cluster_centers.csv"

    def __init__(self, model: nn.Module, cluster_centers):
        super().__init__(model)
        self.cluster_centers = cluster_centers

    def to(self, device: "str | torch.device | int"):
        self.cluster_centers = self.cluster_centers.to(device)
        return super().to(device)

    def forward(self, images: Tensor) -> Tensor:
        embeddings = self.model(images)
        distances = torch.cdist(embeddings, self.cluster_centers)

        prob = torch.reciprocal(distances + 0.01) ** 7
        prob /= prob.sum(1, keepdim=True)

        return prob

    @classmethod
    def load(cls, path: Path | str):
        path = Path(path)
        assert path.is_dir()
        cluster_centers = torch.from_numpy(
            np.loadtxt(
                path / cls.cluster_centers_filename, delimiter=",", dtype=np.float32
            )
        )
        model = ResNet18.from_file(path / cls.model_weights_filename)
        return cls(model, cluster_centers)

    def save(self, path: Path | str) -> None:
        path = Path(path)
        assert path.is_dir()
        np.savetxt(
            path / self.cluster_centers_filename,
            self.cluster_centers.numpy(force=True),
            fmt="%11.5f",
            delimiter=",",
        )
        return super().save(path)


def load_identifier_model(
    path: Path, image_size: Sequence[int] | None = None
) -> IdentifierIdCNN | IdentifierContrastive:
    try:
        model = IdentifierContrastive.load(path)
    except FileNotFoundError:
        logging.info("Contrastive model not found in %s", path)
    else:
        return model

    assert image_size is not None, "Image size must be provided for IdCNN model"
    model = IdentifierIdCNN.load(image_size, path)

    assert model is not None
    return model



import warnings
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import Any

import numpy as np
import torch

from idtrackerai import Blob, ListOfBlobs, ListOfFragments, Session
from idtrackerai.base.network import DEVICE, IdentifierBase, get_onthefly_dataloader
from idtrackerai.utils import conf, create_dir, save_trajectories, track

from .assign_them_all import close_trajectories_gaps
from .correct_impossible_jumps import correct_impossible_velocity_jumps


@torch.inference_mode()
def compute_identity_probabilities(
    identifier_model: IdentifierBase,
    list_of_blobs: ListOfBlobs,
    id_images_file_paths: Sequence[Path],
) -> None:
    blobs_to_predict: list[Blob] = []
    blob_images: list[tuple[int, int]] = []
    identities: list[int] = []
    for blob in list_of_blobs.all_blobs:
        ids = list(blob.final_identities)
        if len(ids) != 1 or ids[0] in (None, 0):
            continue
        try:
            blob_images.append((blob.id_image_index, blob.episode))
        except AttributeError:
            continue
        else:
            identities.append(ids[0])
            blobs_to_predict.append(blob)

    probabilities = []
    identifier_model.to(DEVICE)
    identifier_model.eval()
    dataloader = get_onthefly_dataloader(
        blob_images, id_images_file_paths, np.asarray(identities) - 1
    )
    for images, labels in track(dataloader, "Predicting final identity certainties"):
        probs = identifier_model(images.to(DEVICE))
        probabilities.append(probs[torch.arange(len(labels)), labels].numpy(force=True))

    for blob, probability in zip(blobs_to_predict, np.concatenate(probabilities)):
        blob.identity_certainty = probability


def trajectories_API(
    session: Session,
    list_of_blobs: ListOfBlobs,
    list_of_fragments: ListOfFragments,
    identifier_model: IdentifierBase | None,
) -> None:
    session.velocity_threshold = _get_velocity_threshold(list_of_fragments)

    if (
        session.track_wo_identities
        or session.single_animal
        or len(list_of_fragments) < 2
    ):
        session.estimated_accuracy = 1.0
    else:
        with session.new_timer("Impossible jumps correction"):
            correct_impossible_velocity_jumps(session, list_of_fragments)
            session.individual_fragments_stats = list_of_fragments.get_stats()
            session.estimated_accuracy = compute_estimated_accuracy(list_of_fragments)
            list_of_fragments.update_blobs(list_of_blobs.all_blobs)
        with session.new_timer("Crossings solver"):
            close_trajectories_gaps(session, list_of_blobs, list_of_fragments)

    if identifier_model is not None:
        compute_identity_probabilities(
            identifier_model, list_of_blobs, list_of_fragments.id_images_file_paths
        )

    create_dir(session.trajectories_folder, remove_existing=True)

    output_dict = produce_output_dict(list_of_blobs.blobs_in_video, session)
    save_trajectories(
        session.trajectories_folder, output_dict, session.trajectories_formats
    )


def _get_velocity_threshold(list_of_fragments: ListOfFragments) -> float:
    distances = np.concatenate(
        [
            fragment.frame_by_frame_velocity
            for fragment in list_of_fragments.individual_fragments
        ]
    )
    if distances.size == 0:
        return np.nan
    return float(
        2 * np.percentile(distances, conf.VEL_PERCENTILE, overwrite_input=True)
    )


def produce_output_dict(
    blobs_in_video: list[list[Blob]],
    session: Session,
    progress_bar=None,
    abort: Callable = lambda: False,
) -> dict[str, Any]:
    """Outputs the dictionary with keys: trajectories, git_commit, video_path,
    frames_per_second

    Parameters
    ----------
    blobs_in_video : list
        List of all blob objects (see :class:`Blob`) generated by
        considering all the blobs segmented from the video
    session : <Session object>
        See :class:`~video.Video`

    Returns
    -------
    dict
        Output dictionary containing trajectories as values
    """

    trajectories = np.full((session.number_of_frames, session.n_animals, 2), np.nan)
    id_probabilities = np.full((session.number_of_frames, session.n_animals, 1), np.nan)
    areas = np.full((session.number_of_frames, session.n_animals), np.nan)

    for frame_number, blobs_in_frame in enumerate(
        track(blobs_in_video, "Producing trajectories")
    ):
        if abort():
            return {}
        if progress_bar:
            progress_bar.emit(frame_number)
        for blob in blobs_in_frame:
            for identity, centroid in blob.final_ids_and_centroids:
                if identity not in (None, 0):
                    trajectories[blob.frame_number, identity - 1] = centroid
            blob_final_identities = list(blob.final_identities)

            if (
                blob.is_an_individual
                and len(blob_final_identities) == 1
                and blob_final_identities[0] not in (None, 0)
            ):
                identity = blob_final_identities[0]
                areas[blob.frame_number, identity - 1] = blob.area
                id_probabilities[blob.frame_number, identity - 1] = (
                    blob.identity_certainty
                )
    # id_probabilities for missing trajectories
    id_probabilities[np.isnan(trajectories[..., 0])] = 0

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)  # mean of empty slice
        return {
            "trajectories": trajectories,
            "id_probabilities": id_probabilities,
            "version": session.version,
            "height": session.height,
            "width": session.width,
            "video_paths": list(map(str, session.video_paths)),
            "frames_per_second": session.frames_per_second,
            "body_length": session.median_body_length,
            "estimated_accuracy": session.estimated_accuracy,
            "areas": {
                "mean": np.nanmean(areas, axis=0),
                "median": np.nanmedian(areas, axis=0),
                "std": np.nanstd(areas, axis=0),
            },
            "setup_points": session.setup_points,
            "identities_labels": session.identities_labels
            or [str(i + 1) for i in range(session.n_animals)],
            "identities_groups": {
                key: list(value) for key, value in session.identities_groups.items()
            },
            "length_unit": session.length_unit,
            "silhouette_score": session.silhouette_score,
            "fragment_connectivity": session.fragment_connectivity,
            "fraction_identified": np.mean(np.isfinite(trajectories)),
        }


def compute_estimated_accuracy(list_of_fragments: ListOfFragments) -> float:
    weighted_P2 = 0
    number_of_individual_blobs = 0

    for fragment in list_of_fragments.individual_fragments:
        if fragment.assigned_identities[0] not in (0, None):
            assert fragment.P2_vector is not None
            weighted_P2 += (
                fragment.P2_vector[fragment.assigned_identities[0] - 1]
                * fragment.n_images
            )
        number_of_individual_blobs += fragment.n_images
    return float(weighted_P2 / number_of_individual_blobs)



import json
import logging
from collections.abc import Iterable, Sequence
from datetime import datetime
from importlib import metadata
from itertools import count, pairwise
from math import sqrt
from os import cpu_count
from pathlib import Path
from statistics import fmean
from typing import Any, Literal

import cv2
import h5py
import numpy as np

from . import IdtrackeraiError
from .utils import (
    Episode,
    LengthCalibration,
    Timer,
    assert_all_files_exist,
    assert_knowledge_transfer_is_possible,
    build_ROI_mask_from_list,
    create_dir,
    json_default,
    json_object_hook,
    remove_dir,
    resolve_path,
    track,
)


class Session:
    """
    A class containing the main features of the session.

    This class includes properties of the video by itself, user defined
    parameters for the tracking, and other properties that are generated
    throughout the tracking process.

    We use this class as a storage of data coming from different processes.
    However, this is bad practice and it will change in the future.
    """

    velocity_threshold: float
    erosion_kernel_size: int
    ratio_accumulated_images: float
    individual_fragments_stats: dict
    session_folder: Path
    setup_points: dict[str, list[tuple[int, int]]]
    median_body_length: float
    """median of the diagonals of individual blob's bounding boxes"""
    first_frame_first_global_fragment: int = -1
    identities_groups: dict
    """Named groups of identities stored in the validation GUI.
    If `exclusive ROI`, the identities of each region will be saved here"""
    length_calibrations: list[LengthCalibration]
    """List of length calibrations containing two points (in pixels units) and
    the real distance between these two defined by the user in the Validator."""
    episodes: list[Episode]
    """Indicates the starting and ending frames of each video episode.
    Video episodes are used for parallelization of some processes"""
    width: int
    """Video width in pixels"""
    height: int
    """Video height in pixels"""
    frames_per_second: float
    """Video frame rate in frames per second obtained by OpenCV from the
    video file"""
    accumulation_statistics_data: dict[str, list]
    version: str
    """Version of idtracker.ai"""
    number_of_error_frames: int = -1
    """The number of frames with more blobs than animals. Set on animals_detection."""
    estimated_accuracy: float | None = None
    identities_labels: list[str] | None = None
    """A list with a name for every identity. Defined and used in validator"""
    identities_colors: list[str] | None = None
    """A list with a color for every identity. Defined and used in validator"""
    background_from_segmentation_gui: Path | None = None
    """Path to the background computed by the segmentation app. It is reused at tracking time"""

    video_paths: list[Path] = []
    """List of paths to the different files the video is composed of.
    If the video is a single file, the list will have length 1"""
    number_of_animals: int = 0
    intensity_ths: None | Sequence[float] = None
    area_ths: None | Sequence[float] = None
    # bkg_model: None | np.ndarray = None
    name: str = ""
    output_dir: Path | None | str = None
    tracking_intervals: list | None = None
    resolution_reduction: float | None = None
    roi_list: list[str] | str | None = None
    use_bkg: bool = False
    knowledge_transfer_folder: None | Path = None
    check_segmentation: bool = False
    identity_transfer: bool = False
    track_wo_identities: bool = False
    frames_per_episode: int = 500
    background_subtraction_stat: Literal["median", "mean", "max", "min"] | str = (
        "median"
    )
    number_of_frames_for_background: int = 50
    number_of_parallel_workers: int = 0
    data_policy: Literal[
        "trajectories", "validation", "knowledge_transfer", "idmatcher.ai", "all"
    ] = "idmatcher.ai"
    id_image_size: list[int] = []
    """ Shape of the Blob's identification images (width, height, n_channels)"""
    trajectories_formats: Sequence[Literal["h5", "npy", "csv", "pickle"]] = [
        "h5",
        "npy",
        "csv",
    ]
    """A sequence of strings defining in which formats the trajectories should be saved"""
    exclusive_rois: bool = False
    """Treat each separate ROI as closed identities groups"""
    identity_transfer_succeeded: bool = False
    "True if the identity transfer has been done successfully"
    bounding_box_images_in_ram: bool = False
    "Keep bounding box images on RAM and until used, never write them on disk"
    last_validated: datetime | None = None
    "Last time this session was validated using the Validator"
    silhouette_score: float | None = None
    "Silhouette score reached at the end of the contrastive step"
    fragment_connectivity: float | None = None
    "Connectivity of the fragments used in the contrastive step"

    def set_parameters(self, reset: bool = False, **parameters) -> set[str]:
        """Sets parameters to self only if they are present in the class annotations.

        Parameters
        ----------
        reset : bool, optional
            If True, resets the session parameters before setting new ones, by default False.
        **parameters : dict
            Arbitrary keyword arguments representing the parameters to set.

        Returns
        -------
        set[str]
            The set of non recognized parameters names.
        """
        if reset:
            logging.info("Restarting Session parameters")
            for key in list(self.__dict__.keys()):
                if key in self.__class__.__annotations__:
                    del self.__dict__[key]
        non_recognized_parameters: set[str] = set()
        for param, value in parameters.items():
            lower_param = param.lower()
            if lower_param in self.__class__.__annotations__:
                setattr(self, lower_param, value)
            else:
                non_recognized_parameters.add(param)
        return non_recognized_parameters

    def prepare_tracking(self) -> None:
        """Initializes the session object, checking all parameters"""
        logging.debug("Initializing Session")
        try:
            self.version = metadata.version("idtrackerai")
        except metadata.PackageNotFoundError:
            raise IdtrackeraiError("idtrackerai package is not installed")

        if not isinstance(self.video_paths, list):
            video_paths = [self.video_paths]
        else:
            video_paths = self.video_paths
        self.assert_video_paths(video_paths)
        self.video_paths = [resolve_path(path) for path in video_paths]
        logging.info(
            "Setting video paths to:\n    " + "\n    ".join(map(str, self.video_paths))
        )

        if self.area_ths is None:
            raise IdtrackeraiError("Missing area thresholds parameter")

        if self.intensity_ths is None:
            raise IdtrackeraiError("Missing intensity thresholds parameter")

        if self.knowledge_transfer_folder is not None:
            self.knowledge_transfer_folder = resolve_path(
                self.knowledge_transfer_folder
            )
            if not self.knowledge_transfer_folder.exists():
                raise IdtrackeraiError(
                    f'Knowledge transfer folder "{self.knowledge_transfer_folder}" not'
                    " found"
                )

            if (
                self.knowledge_transfer_folder.is_dir()
                and self.knowledge_transfer_folder.name.startswith("session_")
            ):
                if (self.knowledge_transfer_folder / "accumulation_0").is_dir():
                    self.knowledge_transfer_folder /= "accumulation_0"
                else:
                    self.knowledge_transfer_folder /= "accumulation"

            self.id_image_size, self.resolution_reduction = (
                assert_knowledge_transfer_is_possible(
                    self.knowledge_transfer_folder, self.n_animals
                )
            )

        self.width, self.height, self.frames_per_second = (
            self.get_info_from_video_paths(self.video_paths)
        )
        self.number_of_frames, _, self.tracking_intervals, self.episodes = (
            self.get_processing_episodes(
                self.video_paths, self.frames_per_episode, self.tracking_intervals
            )
        )

        trackable_n_frames = sum(
            interv[1] - interv[0] for interv in self.tracking_intervals
        )
        if trackable_n_frames != self.number_of_frames:
            logging.info(
                f"The session has {self.number_of_frames} frames from which "
                f"{trackable_n_frames} will be tracked "
                f"({self.number_of_episodes} episodes)"
            )
        else:
            logging.info(
                f"The session has {self.number_of_frames} "
                f"frames ({self.number_of_episodes} episodes)"
            )
        if len(self.episodes) < 10:
            for episode in self.episodes:
                video_name = episode.video_path.name
                logging.info(
                    f"\tEpisode {episode.index}, frames ({episode.local_start} "
                    f"-> {episode.local_end}) of /{video_name}"
                )
        assert self.number_of_episodes > 0

        if self.output_dir is None:
            self.output_dir = self.video_paths[0].parent
        self.output_dir = resolve_path(self.output_dir)

        if not self.name:
            self.name = "&".join(path.stem for path in self.video_paths)
            logging.info('No session name provided, assigning name "%s"', self.name)

            if (self.output_dir / f"session_{self.name}").exists():
                # add a counter in sessions with default name
                for index in count(1):
                    name = self.name + f"_{index}"
                    if not (self.output_dir / f"session_{name}").exists():
                        break
                else:
                    raise RuntimeError
                logging.info(
                    'A session with the assigned name ("%s") already exists, renaming'
                    ' current session to "%s"',
                    self.name,
                    name,
                )
                self.name = name

        self.session_folder = self.output_dir / f"session_{self.name}"

        create_dir(self.session_folder)
        create_dir(self.preprocessing_folder)

        self.ROI_mask = build_ROI_mask_from_list(self.roi_list, self.width, self.height)

        if isinstance(self.id_image_size, int):
            self.id_image_size = [self.id_image_size, self.id_image_size, 1]
        elif not self.id_image_size:  # if it is None or empty tuple or list...
            self.id_image_size = []

        if self.number_of_parallel_workers <= 0:
            computer_CPUs = cpu_count()
            if computer_CPUs is not None:
                if self.number_of_parallel_workers == 0:
                    self.number_of_parallel_workers = min((computer_CPUs + 1) // 2, 4)
                elif self.number_of_parallel_workers < 0:
                    self.number_of_parallel_workers += computer_CPUs
            else:
                logging.warning(
                    "Could not determine the number of CPUs in the computer."
                    " Setting the number of parallel jobs to 1"
                )
                self.number_of_parallel_workers = 1
        logging.info("Number of parallel jobs: %d", self.number_of_parallel_workers)

        if self.number_of_animals == 0 and not self.track_wo_identities:
            raise IdtrackeraiError(
                "Cannot track with an undefined number of animals (n_animals = 0)"
                " when tracking with identities"
            )

        if (
            self.background_from_segmentation_gui is not None
            and self.background_from_segmentation_gui.is_file()
        ):
            # If the background was computed by the segmentation GUI, we move it to the final location
            self.background_path.unlink(missing_ok=True)
            self.background_from_segmentation_gui.rename(self.background_path)
            logging.info(
                f"Background from Segmentation App moved to {self.background_path}"
            )

        self.identities_groups = {}
        self.setup_points = {}

        # Processes timers
        self.timers: dict[str, Timer] = {}

    def save(self) -> None:
        """Saves the instantiated Session object"""
        logging.info(f"Saving Session object in {self.path_to_session}", stacklevel=2)
        dict_to_save = (self.defaults() | vars(self)).copy()
        dict_to_save.pop("episodes", None)
        dict_to_save.pop("output_dir", None)
        dict_to_save.pop("background_from_segmentation_gui", None)
        self.path_to_session.write_text(
            json.dumps(dict_to_save, default=json_default, indent=4)
        )

    @classmethod
    def load(
        cls,
        path: Path | str,
        video_paths_dir: Path | str | None = None,
        allow_not_found_video_files: bool = True,
    ) -> "Session":
        """Load a session object stored in a JSON file

        Parameters
        ----------
        path : Path | str
            Path to the session JSON file to load from or to the session folder containing the JSON file.
        video_paths_dir : Path | None, optional
            Folder containing the video paths. If None (the default), they are expected to be in:

            - In the same location where they were during the tracking
            - In the parent folder of the JSON file being loaded
            - In the double-parent folder of the JSON file being loaded
            - In the parent folder of the original path of the JSON file
            - In the double-parent folder of the original path of the JSON file
            - In the current working directory

        allow_not_found_video_files : bool, optional
            If False, an IdtrackeraiError exception is raised if the video files couldn't be found, by default True.

        Returns
        -------
        Session
            Instance of :class:`Session` from the loaded JSON file.

        Raises
        ------
        FileNotFoundError
            If the JSON couldn't be found.
        IdtrackeraiError
            If the video files couldn't be found and `allow_not_found_video_files` is `True`.
        ValueError
            If the input file is not JSON readable.
            If the session being loaded is from version 4 or lower of idtracker.ai.
        """
        path = resolve_path(path)
        logging.info(f"Loading Session from {path}", stacklevel=2)
        if not path.exists():
            raise FileNotFoundError(f"{path} not found")
        if not path.is_file():
            possible_files = ("session.json", "video_object.json", "video_object.npy")
            for file in possible_files:
                if (path / file).is_file():
                    path /= file
                    break
            else:
                raise FileNotFoundError(
                    f"Session parameterfiles not found in {path}"
                    if path.name.startswith("session_")
                    else f"The path {path} is not a session folder, session folders start with 'session_'"
                )

        if path.suffix == ".npy":
            raise ValueError(
                f"The file {path} is no longer supported by idtrackerai "
                "v6 (or higher) since it was generated by v4 (or lower)"
            )

        # Avoid loading huge video files by mistake
        if path.stat().st_size > 5000000:  # 5MB
            raise ValueError(
                f"{path} takes {path.stat().st_size / (1024**2):.1f} MB, it does not seem"
                " like a session json file"
            )

        try:
            with open(path, encoding="utf_8") as file:
                session_dict: dict[str, Any] = json.load(
                    file, object_hook=json_object_hook
                )
        except UnicodeDecodeError as exc:
            raise ValueError(
                f'The file "{path}" is not UTF-8 readable. It should be a '
                f'"session.json" file or an entire session folder. Original error message: {exc}'
            ) from exc
        except json.JSONDecodeError as exc:
            raise ValueError(
                f'The file "{path}" is not JSON readable. Original JSON error message: {exc}'
            ) from exc
        if "original_width" in session_dict:
            session_dict["width"] = session_dict["original_width"]

        if "original_height" in session_dict:
            session_dict["height"] = session_dict["original_height"]

        if "n_animals" not in session_dict and "number_of_animals" in session_dict:
            session_dict["n_animals"] = session_dict["number_of_animals"]

        session_dict["video_paths"] = [
            resolve_path(pth) for pth in session_dict["video_paths"]
        ]

        if "session" in session_dict and "name" not in session_dict:
            # backward compatibility
            session_dict["name"] = session_dict.pop("session")

        session_dict["timers"] = {
            name: Timer.from_dict(timer_dict)
            for name, timer_dict in session_dict.get("timers", {}).items()
        }

        # format timers and Paths
        for key, value in session_dict.items():
            if key.endswith("_timer") and isinstance(value, dict):
                # <=5.2.11 compatibility
                session_dict[key] = Timer.from_dict(value)
            if key.endswith("_folder") and isinstance(value, str):
                session_dict[key] = resolve_path(value)

        if "general_timer" in session_dict:
            # This is the only timer we currently use after tracking,
            # so we want to recover it if it was saved in a previous version style
            session_dict["timers"]["Tracking session"] = session_dict["general_timer"]

        if session_dict.get("length_calibrations"):
            session_dict["length_calibrations"] = [
                LengthCalibration.from_dict(value)
                for value in session_dict["length_calibrations"]
            ]

        session_dict["last_validated"] = (
            datetime.fromisoformat(session_dict["last_validated"])
            if session_dict.get("last_validated") is not None
            else None
        )

        session = cls.__new__(cls)
        session.__dict__.update(session_dict)

        try:
            session.update_paths(path.parent, video_paths_dir)
        except FileNotFoundError as exc:
            if not allow_not_found_video_files:
                # We raise an IdtrackeraiError to distinguish from the
                # FileNotFoundError raised when the .json file is not found
                raise IdtrackeraiError() from exc

        try:
            _, _, _, session.episodes = session.get_processing_episodes(
                session.video_paths,
                session.frames_per_episode,
                session.tracking_intervals,
            )
        except AttributeError:
            logging.warning(
                "Could not load video episodes probably due to loading an old version"
                " session"
            )
        except (IdtrackeraiError, FileNotFoundError) as exc:
            logging.warning("Could not load video episodes. %s", str(exc))

        return session

    def new_timer(self, name: str) -> Timer:
        """Generates, saves and returns a Timer"""
        timer = Timer(name)
        self.timers[name] = timer
        return timer

    def __str__(self) -> str:
        return f"<session {self.session_folder}>"

    def set_id_image_size(self, median_body_length: float) -> None:
        """Sets the :attr:`median_body_length` and computes the appropiate identification image size and resolution reduction when needed and taking into account the existing user defined values.

        Parameters
        ----------
        median_body_length : float
            Median body length of all individual blobs in the video
        """
        max_size = 80  # has to be even
        self.median_body_length = median_body_length
        auto_size = int(median_body_length / sqrt(2) + 0.5)
        auto_size += auto_size % 2  # nearest even integer
        logging.info(f"The automatic identification image size is {auto_size}")

        if not self.id_image_size:
            if self.resolution_reduction is None:
                if auto_size > max_size:
                    self.resolution_reduction = np.round(max_size / auto_size, 2)
                    self.id_image_size = [max_size, max_size, 1]
                    logging.info(
                        f"Since this is bigger than {max_size}, the resolution "
                        f"reduction is set to {self.resolution_reduction} to diminish"
                        f" this image size to {max_size} pixels"
                    )
                else:
                    logging.info("No resolution reduction required")
                    self.resolution_reduction = 1
                    self.id_image_size = [auto_size, auto_size, 1]
            else:
                scaled_size = int(auto_size * self.resolution_reduction + 0.5)
                scaled_size += scaled_size % 2
                logging.info(
                    f"The resolution reduction is fixed to {self.resolution_reduction}"
                    " so the automatic identification image size is rescaled "
                    f"to {scaled_size}"
                )
                self.id_image_size = [scaled_size, scaled_size, 1]
        else:
            logging.info(
                f"But the identification image size is fixed by the user to {self.id_image_size}"
            )
            if self.resolution_reduction is None:
                if auto_size > self.id_image_size[0]:
                    self.resolution_reduction = np.round(
                        self.id_image_size[0] / auto_size, 2
                    )
                    logging.info(
                        f"Since animals are bigger than the image size defined by the"
                        " user, the resolution reduction is set to "
                        f"{self.resolution_reduction} to fit them"
                    )
                else:
                    logging.info("No resolution reduction required")
                    self.resolution_reduction = 1
            else:
                logging.info(
                    "The identification image size and the resolution "
                    "reduction are both fixed by the user."
                )

        logging.info(
            f"Identification image size set to {self.id_image_size},"
            f" resolution reduction factor set to {self.resolution_reduction}"
        )
        if self.id_image_size[0] > max_size:
            logging.warning(
                f"The identification image size is bigger than {max_size}. "
                "This can unnecessarily slow down the models training"
            )

    @property
    def n_animals(self) -> int:
        return self.number_of_animals

    @property
    def single_animal(self) -> bool:
        return self.n_animals == 1

    @property
    def bkg_model(self) -> np.ndarray | None:
        if self.background_path.is_file():
            # Use fromfile to handle paths with non-ASCII characters
            return cv2.imdecode(  # type: ignore
                np.fromfile(self.background_path, dtype=np.uint8), cv2.IMREAD_COLOR
            )[..., 0]
        return None

    @bkg_model.setter
    def bkg_model(self, bkg: np.ndarray | None) -> None:
        if bkg is None:
            del self.bkg_model
            return
        # cv2.imwrite has given issues with paths containing chinese characters
        cv2.imencode(self.background_path.suffix, bkg)[1].tofile(self.background_path)
        logging.info(f"Background saved at {self.background_path}")

    @bkg_model.deleter
    def bkg_model(self) -> None:
        self.background_path.unlink(missing_ok=True)

    @property
    def ROI_mask(self) -> np.ndarray | None:
        if self.ROI_mask_path.is_file():
            return cv2.imdecode(  # type: ignore
                np.fromfile(self.ROI_mask_path, dtype=np.uint8), cv2.IMREAD_COLOR
            )[..., 0]
        return None

    @ROI_mask.setter
    def ROI_mask(self, mask: np.ndarray | None) -> None:
        if mask is None:
            del self.ROI_mask
            return
        # cv2.imwrite has given issues with paths containing chinese characters
        cv2.imencode(self.ROI_mask_path.suffix, mask)[1].tofile(self.ROI_mask_path)
        logging.info(f"ROI mask saved at {self.ROI_mask_path}")

    @ROI_mask.deleter
    def ROI_mask(self) -> None:
        self.ROI_mask_path.unlink(missing_ok=True)

    @property
    def number_of_episodes(self) -> int:
        "Number of episodes in which the video is splitted for parallel processing"
        return len(self.episodes)

    # Paths and folders
    @property
    def preprocessing_folder(self) -> Path:
        return self.session_folder / "preprocessing"

    @property
    def background_path(self) -> Path:
        return self.preprocessing_folder / "background.png"

    @property
    def ROI_mask_path(self) -> Path:
        return self.preprocessing_folder / "ROI_mask.png"

    @property
    def trajectories_folder(self) -> Path:
        return self.session_folder / "trajectories"

    @property
    def crossings_detector_folder(self) -> Path:
        return self.session_folder / "crossings_detector"

    @property
    def individual_videos_folder(self) -> Path:
        return self.session_folder / "individual_videos"

    @property
    def accumulation_folder(self) -> Path:
        return self.session_folder / "accumulation"

    @property
    def id_images_folder(self) -> Path:
        return self.session_folder / "identification_images"

    @property
    def blobs_path(self) -> Path:
        """get the path to save the blob collection after segmentation.
        It checks that the segmentation has been successfully performed"""
        return self.preprocessing_folder / "list_of_blobs.pickle"

    @property
    def global_fragments_path(self) -> Path:
        """get the path to save the list of global fragments after
        fragmentation"""
        return self.preprocessing_folder / "list_of_global_fragments.json"

    @property
    def fragments_path(self) -> Path:
        """get the path to save the list of global fragments after
        fragmentation"""
        return self.preprocessing_folder / "list_of_fragments.json"

    @property
    def path_to_session(self) -> Path:
        return self.session_folder / "session.json"

    @property
    def bbox_images_folder(self) -> Path:
        return self.session_folder / "bounding_box_images"

    @property
    def id_images_file_paths(self) -> list[Path]:
        # From version 6.0.0 id_images are saved using the .h5 extension, not .hdf5
        extension = "hdf5" if int(self.version.split(".")[0]) < 6 else "h5"
        try:
            return [
                self.id_images_folder / f"id_images_{e}.{extension}"
                for e in range(self.number_of_episodes)
            ]
        except AttributeError:
            # Loading a Session without the video files being present generates a session
            # without episodes. In this case, lets take all present files in id_images_folder
            paths: list[Path] = []
            for episode in count():
                path = self.id_images_folder / f"id_images_{episode}.{extension}"
                if not path.exists():
                    return paths
                paths.append(path)
            else:
                raise  # for PyLance

    @classmethod
    def defaults(cls) -> dict[str, Any]:
        return {
            key: value
            for key, value in vars(cls).items()
            if not key.startswith("__")
            and not callable(value)
            and not callable(getattr(value, "__get__", None))
        }

    def update_paths(
        self, new_session_path: Path, user_video_paths_dir: Path | str | None = None
    ) -> None:
        """Update paths of objects (e.g. blobs_path, preprocessing_folder...)
        according to the location of the new Session object given
        by `new_session_path`.
        """
        logging.info(
            f"Searching video files: {[str(path.name) for path in self.video_paths]}"
        )
        folder_candidates: set[Path] = {
            self.video_paths[0],
            new_session_path,
            new_session_path.parent,
            self.session_folder.parent,
            self.session_folder,
            Path.cwd(),
        }
        if user_video_paths_dir is not None:
            folder_candidates.add(Path(user_video_paths_dir))

        need_to_save = False
        if self.session_folder != new_session_path:
            logging.info(
                f"Updated session folder from {self.session_folder} to"
                f" {new_session_path}"
            )
            self.session_folder = new_session_path
            need_to_save = True

        for folder_candidate in folder_candidates:
            if folder_candidate is None:
                continue

            try:
                # This can raise PermissionError if the path is not accessible
                if folder_candidate.is_file():
                    folder_candidate = folder_candidate.parent

                candidate_new_video_paths = [
                    folder_candidate / path.name for path in self.video_paths
                ]

                assert_all_files_exist(candidate_new_video_paths)
            except (FileNotFoundError, PermissionError):
                continue

            logging.info("All video files found in %s", folder_candidate)
            break
        else:
            raise FileNotFoundError(
                "Video file paths not found:\n    "
                + "\n    ".join(str(p) for p in self.video_paths)
            )

        if self.video_paths != candidate_new_video_paths:
            logging.info("Updating new video files paths")
            self.video_paths = candidate_new_video_paths
            need_to_save = True

        if need_to_save:
            self.save()

    @staticmethod
    def assert_video_paths(video_paths: Iterable[Path | str]) -> None:
        if not video_paths:
            raise IdtrackeraiError("Empty Video paths list")

        for path in video_paths:
            path = resolve_path(path)

            try:
                if path.is_dir():
                    raise IdtrackeraiError(
                        f'Directories ("{path}") are not allowed as video file paths'
                    )
            except PermissionError:
                raise IdtrackeraiError(f'Permission denied for the path "{path}"')

            if not path.is_file():
                raise IdtrackeraiError(f'Video file "{path}" not found')

            readable = cv2.VideoCapture(str(path)).grab()
            if not readable:
                raise IdtrackeraiError(f'Video file "{path}" not readable by OpenCV.')

    @staticmethod
    def get_info_from_video_paths(
        video_paths: Iterable[Path | str],
    ) -> tuple[int, int, float]:
        """Gets some information about the video from the video file itself.

        Returns:
            width: int, height: int, fps: float
        """

        widths, heights, fps = [], [], []
        for path in video_paths:
            cap = cv2.VideoCapture(str(path))
            widths.append(int(cap.get(cv2.CAP_PROP_FRAME_WIDTH)))
            heights.append(int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)))

            try:
                fps.append(cap.get(cv2.CAP_PROP_FPS))
            except cv2.error:
                logging.warning(f"Cannot read frame per second for {path}")
                fps.append(None)
            cap.release()

        if len(set(widths)) != 1 or len(set(heights)) != 1:
            raise IdtrackeraiError("Video paths have different resolutions")

        if len(set(fps)) != 1:
            fps = [float(np.mean(fps))]
            logging.warning(
                f"Different frame rates detected ({fps}). "
                f"Setting the frame rate to the mean value: {fps[0]} fps"
            )

        return widths[0], heights[0], fps[0]

    @staticmethod
    def get_processing_episodes(
        video_paths: Sequence[Path | str],
        frames_per_episode: int,
        tracking_intervals=None,
    ) -> tuple[(int, list[int], list[list[int]], list[Episode])]:
        """Process the episodes by getting the number of frames in each video
        path and the tracking interval.

        Episodes are used to compute processes in parallel for different
        parts of the video. They are a tuple with

        - local start frame
        - local end frame
        - video path index
        - global start frame
        - global end frame

        where "local" means in the specific video path and "global" means in
        the whole (multi path) video

        Episodes are guaranteed to belong to a single video path and to have
        all of their frames (end not included) inside a the tracking interval
        """

        def in_which_interval(frame_number, intervals) -> int | None:
            for i, (start, end) in enumerate(intervals):
                if start <= frame_number < end:
                    return i
            return None

        for path in video_paths:
            if not Path(path).exists():
                raise FileNotFoundError(f"{path} not found")

        # total number of frames for every video path
        video_paths_n_frames = [
            int(cv2.VideoCapture(str(path)).get(cv2.CAP_PROP_FRAME_COUNT))
            for path in video_paths
        ]

        for n_frames, video_path in zip(video_paths_n_frames, video_paths):
            if n_frames <= 0:
                raise IdtrackeraiError(
                    f"OpenCV cannot read the number of frames in {video_path}"
                )
        number_of_frames = sum(video_paths_n_frames)

        # set full tracking interval if not defined
        if tracking_intervals is None:
            tracking_intervals = [[0, number_of_frames]]
        elif isinstance(tracking_intervals[0], int):
            tracking_intervals = [tracking_intervals]

        # find the global frames where the video path changes
        video_paths_changes = [0] + list(np.cumsum(video_paths_n_frames))

        # build an interval list like ("frame" refers to "global frame")
        #   [[first frame of video path 0, last frame of video path 0],
        #    [first frame of video path 1, last frame of video path 1],
        #    [...]]
        video_paths_intervals = list(pairwise(video_paths_changes))

        # find the frames where a tracking interval starts or ends
        tracking_intervals_changes = list(np.asarray(tracking_intervals).ravel())

        # Take into account tracking interval changes
        # and video path changes to compute episodes
        limits = video_paths_changes + tracking_intervals_changes

        # clean repeated limits and sort them
        limits = sorted(set(limits))

        # Create "long episodes" as the intervals between any video path
        # change or tracking interval change (keeping only the ones that
        # are inside a tracking interval)
        long_episodes = []
        for start, end in pairwise(limits):
            if (
                in_which_interval(start, tracking_intervals) is not None
            ) and 0 <= start < number_of_frames:
                long_episodes.append((start, end))

        # build definitive episodes by dividing long episodes to fit in
        # the FRAMES_PER_EPISODE restriction
        index = 0
        episodes = []
        for start, end in long_episodes:
            video_path_index = in_which_interval(start, video_paths_intervals)
            assert video_path_index is not None
            global_local_offset = video_paths_intervals[video_path_index][0]

            n_subepisodes = int((end - start) / (frames_per_episode + 1))
            new_episode_limits = np.linspace(start, end, n_subepisodes + 2, dtype=int)
            for new_start, new_end in pairwise(new_episode_limits):
                episodes.append(
                    Episode(
                        index=index,
                        local_start=new_start - global_local_offset,
                        local_end=new_end - global_local_offset,
                        video_path=resolve_path(video_paths[video_path_index]),
                        global_start=new_start,
                        global_end=new_end,
                    )
                )
                index += 1
        return number_of_frames, video_paths_n_frames, tracking_intervals, episodes

    @staticmethod
    def in_which_interval(frame_number, intervals) -> int | None:
        for i, (start, end) in enumerate(intervals):
            if start <= frame_number < end:
                return i
        return None

    @property
    def length_unit(self) -> float | None:
        """Length calibration factor for translating pixel units to user defined units. Property set in the Validator. Returns None if there are no calibrations."""
        if not hasattr(self, "length_calibrations"):
            return None

        values = []
        for c in self.length_calibrations:
            value = c.value()
            if value is not None:
                values.append(value)
        if not values:
            return None
        return fmean(values)

    def delete_data(self) -> None:
        """Deletes some folders with data, to make the outcome lighter.

        Which folders are deleted depends on the constant DATA_POLICY
        """
        logging.info(f"Data policy: {self.data_policy!r}")
        data_policy = self.data_policy.lower()

        if data_policy == "trajectories":
            remove_dir(self.bbox_images_folder)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
            remove_dir(self.accumulation_folder)
            remove_dir(self.preprocessing_folder)
        elif data_policy == "validation":
            remove_dir(self.bbox_images_folder)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
            remove_dir(self.accumulation_folder)
        elif data_policy == "knowledge_transfer":
            remove_dir(self.bbox_images_folder)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
        elif data_policy == "idmatcher.ai":
            remove_dir(self.bbox_images_folder)
            remove_dir(self.crossings_detector_folder)
        elif data_policy == "all":
            pass
        else:
            logging.error(
                "Data Policy is not valid. It has to be one of "
                f'{("trajectories", "validation", "knowledge_transfer", "idmatcher.ai", "all")}'
            )

    def compress_data(self) -> None:
        """Compress the identification images h5py files"""
        if not self.id_images_folder.exists():
            return

        tmp_path = self.session_folder / "tmp.h5py"

        for path in track(
            self.id_images_file_paths, "Compressing identification images"
        ):
            if not path.is_file():
                continue
            with (
                h5py.File(path, "r") as original_file,
                h5py.File(tmp_path, "w") as compressed_file,
            ):
                for key, data in original_file.items():
                    compressed_file.create_dataset(
                        key, data=data, compression="gzip" if "image" in key else None
                    )
            path.unlink()  # Windows needs this call before rename()
            tmp_path.rename(path)



"""isort:skip_file"""

from torch.backends import cudnn

from .device import DEVICE
from .models import (
    IdCNN,
    IdentifierBase,
    IdentifierIdCNN,
    IdentifierContrastive,
    ResNet18,
    load_identifier_model,
)
from .train import (
    evaluate,
    evaluate_only_acc,
    StopTraining,
    train_loop,
    ImageDataset,
    get_dataloader,
    get_predictions,
    get_onthefly_dataloader,
    DataLoaderWithLabels,
)

cudnn.benchmark = True  # make it train faster

__all__ = [
    "DataLoaderWithLabels",
    "evaluate",
    "load_identifier_model",
    "IdentifierIdCNN",
    "DEVICE",
    "IdCNN",
    "ResNet18",
    "evaluate_only_acc",
    "StopTraining",
    "train_loop",
    "ImageDataset",
    "get_dataloader",
    "IdentifierContrastive",
    "get_predictions",
    "get_onthefly_dataloader",
    "IdentifierBase",
]



import logging

import cv2
import numpy as np

from idtrackerai import Blob, Session


def compute_erosion_disk(blobs_in_video: list[list[Blob]]) -> int:
    min_frame_distance_transform = []
    for blobs_in_frame in blobs_in_video:
        if blobs_in_frame:
            min_frame_distance_transform.append(
                compute_min_frame_distance_transform(blobs_in_frame)
            )

    return np.ceil(np.nanmedian(min_frame_distance_transform)).astype(int)
    # return np.ceil(np.nanmedian([compute_min_frame_distance_transform(video, blobs_in_frame)
    #                              for blobs_in_frame in blobs_in_video
    #                              if len(blobs_in_frame) > 0])).astype(np.int)


def compute_min_frame_distance_transform(blobs_in_frame: list[Blob]) -> float:
    max_distance_transform = []
    for blob in blobs_in_frame:
        if blob.is_an_individual:
            try:
                max_distance_transform.append(
                    cv2.distanceTransform(
                        blob.get_bbox_mask(), cv2.DIST_L2, cv2.DIST_MASK_PRECISE
                    ).max()
                )
            except cv2.error:
                logging.warning("Could not compute distance transform for this blob")
    return np.min(max_distance_transform) if max_distance_transform else np.nan


def get_eroded_blobs(
    session: Session, blobs_in_frame: list[Blob], frame_number: int
) -> list[Blob]:
    segmented_frame = np.zeros(
        (int(session.height + 0.5), int(session.width + 0.5)), np.uint8
    )

    for blob in blobs_in_frame:
        segmented_frame = cv2.fillPoly(segmented_frame, [blob.contour], [255])

    segmented_eroded_frame = cv2.erode(
        src=segmented_frame,
        kernel=np.ones(session.erosion_kernel_size, np.uint8),
        iterations=1,
    )

    # Extract blobs info
    contours = cv2.findContours(
        segmented_eroded_frame, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_TC89_KCOS
    )[0]

    return [
        Blob(contour, frame_number=frame_number, pixels_are_from_eroded_blob=True)
        for contour in contours
    ]



from collections.abc import Sequence

from qtpy.QtCore import Qt, Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtWidgets import QComboBox, QHBoxLayout, QLabel, QWidget

from idtrackerai.GUI_tools import InvertibleSlider, LabelRangeSlider


class IntensityThresholds(QWidget):
    newValue = Signal(tuple)

    def __init__(self, parent, min, max):
        super().__init__()
        self.parent_widget = parent
        self.label_nobkg = QLabel("Blob intensity\nthresholds")
        self.label_nobkg.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.label_yesbkg = QLabel("Background\ndifference threshold")
        self.label_yesbkg.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.label_yesbkg.setVisible(False)
        self.range_slider = LabelRangeSlider(parent=parent, min=min, max=max)
        self.simple_slider = InvertibleSlider(min=min, max=max)
        self.range_slider.setVisible(False)
        self.th_type = QComboBox()
        self.th_type.addItems(["Dark animals", "Bright animals", "Custom"])
        self.th_type.currentTextChanged.connect(self.threshold_type_changed)
        self.simple_slider.valueChanged.connect(
            lambda: self.newValue.emit(self.value())
        )
        self.range_slider.valueChanged.connect(self.newValue.emit)
        layout = QHBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.label_nobkg)
        layout.addWidget(self.label_yesbkg)
        layout.addWidget(self.th_type)
        layout.addWidget(self.range_slider)
        layout.addWidget(self.simple_slider)

    def threshold_type_changed(self, new_type: str):
        if new_type == "Dark animals":
            self.range_slider.setVisible(False)
            self.simple_slider.setVisible(True)
            self.simple_slider.set_inverted(False)
        elif new_type == "Bright animals":
            self.range_slider.setVisible(False)
            self.simple_slider.setVisible(True)
            self.simple_slider.set_inverted(True)
        elif new_type == "Custom":
            self.simple_slider.setVisible(False)
            self.range_slider.setVisible(True)
        elif new_type == "using_background":
            self.range_slider.setVisible(False)
            self.simple_slider.setVisible(True)
            self.simple_slider.set_inverted(True)
        else:
            raise ValueError(new_type)
        self.newValue.emit(self.value())

    def bkg_changed(self, bkg):  # bkg is np.ndarray of None
        if bkg is None:
            self.label_nobkg.setVisible(True)
            self.label_yesbkg.setVisible(False)
            self.th_type.setVisible(True)
            self.threshold_type_changed(self.th_type.currentText())
        else:
            self.label_nobkg.setVisible(False)
            self.label_yesbkg.setVisible(True)
            self.th_type.setVisible(False)
            self.threshold_type_changed("using_background")

    def setValue(self, value: Sequence[float] | None):
        if value is None:
            value = (0, 130)  # default value
        value = (max(value[0], 0), min(value[1], 255))
        if value[0] == self.simple_slider.min:
            self.simple_slider.set_value(int(value[1]))
            self.th_type.setCurrentText("Dark_animals")
        elif value[1] == self.simple_slider.max:
            self.simple_slider.set_value(int(value[0]))
            self.th_type.setCurrentText("Bright animals")
        else:
            self.range_slider.setValue((int(value[0]), int(value[1])))
            self.th_type.setCurrentText("Custom")

    def value(self) -> tuple[int, int]:
        if self.range_slider.isVisible():
            return self.range_slider.value()
        return (
            (self.simple_slider.value(), self.simple_slider.max)
            if self.simple_slider.inverted
            else (self.simple_slider.min, self.simple_slider.value())
        )

    def setToolTips(self, tooltip_nobkg: str, tooltip_yesbkg: str):
        self.label_nobkg.setToolTip(tooltip_nobkg)
        self.th_type.setToolTip(tooltip_nobkg)
        self.label_yesbkg.setToolTip(tooltip_yesbkg)
        self.range_slider.setToolTip(tooltip_nobkg)
        # self.background_slider.setToolTip(tooltip_yesbkg)



import logging
import pickle
from collections.abc import Iterable, Iterator, Sequence
from io import BytesIO
from itertools import chain, pairwise, product
from multiprocessing import Pool
from pathlib import Path

import h5py
import numpy as np

from . import Blob
from .utils import (
    LOGGING_QUEUE,
    Episode,
    clean_attrs,
    deprecated,
    open_track,
    resolve_path,
    setup_logging_queue,
    track,
)


class _chain(chain):
    """A re-implementation of itertools.chain with added __len__ method for pretty progress bars"""

    len: int | None

    def __len__(self) -> int | None:
        return self.len

    @classmethod
    def from_iterable_of_sequences(cls, iterable: Iterable[Sequence]) -> Iterator:
        chain = cls.from_iterable(iterable)
        chain.len = sum(map(len, iterable))  # type: ignore (not sure how this should be coded)
        return chain


class ListOfBlobs:
    """Contains all the instances of the class :class:`Blob`."""

    blobs_in_video: list[list[Blob]]
    """Main and only attribute of the class, it contains a list of all
    :class:`Blob` instances in a frame for every frame in the video"""

    def __init__(self, blobs_in_video: list[list[Blob]]):
        logging.info("Generating ListOfBlobs object")
        self.blobs_in_video = blobs_in_video

    @classmethod
    def load(cls, file_path: Path | str, verbose: bool = True) -> "ListOfBlobs":
        """Loads an instance of a class saved in a .pickle file.

        Parameters
        ----------
        file_path : Path | str
            Path to a saved instance of a ListOfBlobs object

        Returns
        -------
        ListOfBlobs
        """
        file_path = resolve_path(file_path)
        logging.info(f"Loading ListOfBlobs from {file_path}", stacklevel=2)
        if not file_path.is_file():
            v4_path = file_path.with_name(
                file_path.name.replace("list_of_blobs", "blobs_collection")
            ).with_suffix(".npy")

            if not v4_path.is_file():
                raise FileNotFoundError(file_path)
            else:
                file_path = v4_path

        if file_path.suffix == ".npy":
            list_of_blobs = cls._load_from_v4(file_path, verbose=verbose)
        else:
            with open_track(file_path, "rb", verbose=verbose) as file:
                list_of_blobs: ListOfBlobs = pickle.load(file)
        list_of_blobs.reconnect()
        return list_of_blobs

    def save(self, file_path: Path | str) -> None:
        """Saves instance of the class with Python's pickle protocol

        Parameters
        ----------
        file_path : Path | str
            Path where to save the object
        """
        file_path = resolve_path(file_path)
        logging.info(f"Saving ListOfBlobs at {file_path}", stacklevel=2)
        file_path.parent.mkdir(exist_ok=True)
        self.disconnect()

        for blob in self.all_blobs:
            clean_attrs(blob)

        with open_track(file_path, "wb") as file:
            pickle.dump(self, file, protocol=pickle.HIGHEST_PROTOCOL)
        self.reconnect()

    @classmethod
    def _load_from_v4(cls, path: Path, verbose: bool = True) -> "ListOfBlobs":
        logging.info("Loading from v4 file: %s", path)
        with open_track(path, "rb", verbose=verbose) as file:
            list_of_blobs: "ListOfBlobs" = np.load(file, allow_pickle=True).item()

        for blob in track(
            list_of_blobs.all_blobs, "Updating objects from an old idtracker.ai version"
        ):
            blob.is_an_individual = blob._is_an_individual  # type:ignore
            blob.fragment_identifier = blob._fragment_identifier  # type:ignore
            blob.identity = blob._identity  # type:ignore
            blob.identity_corrected_solving_jumps = (
                blob._identity_corrected_solving_jumps  # type:ignore
            )
            blob.identities_corrected_closing_gaps = (
                blob._identities_corrected_closing_gaps  # type:ignore
            )
        return list_of_blobs

    @property
    def all_blobs(self) -> Iterator[Blob]:
        "A flattened view of all Blobs in the video"
        return _chain.from_iterable_of_sequences(self.blobs_in_video)

    @property
    def number_of_blobs(self) -> int:
        "Total number of Blobs in self"
        return sum(map(len, self.blobs_in_video))

    @property
    def number_of_crossing_blobs(self) -> int:
        return sum(blob.is_a_crossing for blob in self.all_blobs)

    @property
    def number_of_frames(self) -> int:
        "Number of frames in the video, equal to the number of lists of Blobs in self"
        return len(self.blobs_in_video)

    def __len__(self) -> int:
        return len(self.blobs_in_video)

    def compute_overlapping_between_subsequent_frames(self) -> None:
        """Computes overlapping between blobs in consecutive frames.

        Two blobs in consecutive frames overlap if the intersection of the list
        of pixels of both blobs is not empty.

        See Also
        --------
        :meth:`Blob.overlaps_with`
        :meth:`Blob.now_points_to`
        """
        for blob in self.all_blobs:
            blob.next = ()
            blob.previous = ()

        for blobs, blobs_next in pairwise(
            track(self.blobs_in_video, "Connecting blobs")
        ):
            for blob, blob_next in product(blobs, blobs_next):
                if blob.overlaps_with(blob_next):
                    blob.now_points_to(blob_next)

    def disconnect(self):
        for blob in self.all_blobs:
            blob.next = ()

    def reconnect(self):
        try:
            first_blob = next(self.all_blobs)
        except StopIteration:
            logging.warning("No blobs to reconnect")
            return
        if isinstance(first_blob.next, list):
            logging.info("Converting ListOfBlobs from version older than 5.2.2")
            for blob in self.all_blobs:
                blob.previous = tuple(blob.previous)
                blob.next = ()

        for blob in self.all_blobs:
            for prev_blob in blob.previous:
                prev_blob.next = prev_blob.next + (blob,)

    def set_images_for_identification(
        self,
        episodes: list[Episode],
        id_images_files: list[Path],
        id_image_size: list[int],
        resolution_reduction: float,
        n_jobs: int,
    ) -> None:
        """Computes and saves the images used to classify blobs as crossings
        and individuals and to identify the animals along the video.
        """
        if not (0 < resolution_reduction <= 1):
            logging.warning(
                "Resolution reduction should be ranged within (0, 1], current value is %s",
                resolution_reduction,
            )

        inputs = [
            (
                id_image_size[0],
                id_images_file,
                resolution_reduction,
                episode,
                self.blobs_in_video[episode.global_start : episode.global_end],
            )
            for id_images_file, episode in zip(id_images_files, episodes)
        ]

        if n_jobs == 1:
            for input in track(inputs, "Setting images for identification"):
                self.set_id_images_per_episode(input)
        else:
            with Pool(
                n_jobs,
                maxtasksperchild=3,
                initializer=setup_logging_queue,
                initargs=(LOGGING_QUEUE,),
            ) as p:
                for _ in track(
                    p.imap_unordered(self.set_id_images_per_episode, inputs),
                    "Setting images for identification",
                    len(inputs),
                ):
                    pass

        for input in inputs:
            episode, blobs_in_episode = input[3:]
            if isinstance(episode.bbox_images, BytesIO):
                episode.bbox_images.close()
            for index, blob in enumerate(chain.from_iterable(blobs_in_episode)):
                blob.id_image_index = index
                blob.episode = episode.index

    @staticmethod
    def set_id_images_per_episode(
        inputs: tuple[int, Path, float, Episode, list[list[Blob]]],
    ) -> None:
        id_image_size, file_path, res_reduct, episode, blobs_in_episode = inputs

        with (
            h5py.File(file_path, "w") as id_images_file,
            h5py.File(episode.bbox_images, "r") as bbox_images_file,
        ):
            imgs_to_save = id_images_file.create_dataset(
                "id_images",
                (sum(map(len, blobs_in_episode)), id_image_size, id_image_size),
                np.uint8,
            )

            for index, blob in enumerate(chain.from_iterable(blobs_in_episode)):
                bbox_image: np.ndarray = bbox_images_file[blob.bbox_img_id][:]  # type: ignore
                imgs_to_save[index] = blob.get_image_for_identification(
                    id_image_size, bbox_image, res_reduct
                )

    def remove_centroid(self, frame_number: int, centroid_to_remove, id_to_remove):
        for blob in self.blobs_in_video[frame_number]:
            for indx, (id, centroid) in enumerate(
                zip(blob.all_final_identities, blob.all_final_centroids)
            ):
                if id == id_to_remove:
                    dist = (centroid[0] - centroid_to_remove[0]) ** 2 + (
                        centroid[1] - centroid_to_remove[1]
                    ) ** 2
                    if dist < 1:  # it is the same centroid
                        blob.init_validator_variables()
                        blob.user_generated_centroids[indx] = (-1, -1)
                        blob.user_generated_identities[indx] = -1

    def update_centroid(
        self, frame_number: int, centroid_id: int, old_centroid, new_centroid
    ):
        old_centroid = tuple(old_centroid)
        new_centroid = tuple(new_centroid)
        assert len(old_centroid) == 2
        assert len(new_centroid) == 2
        blobs_in_frame = self.blobs_in_video[frame_number]
        assert blobs_in_frame

        dist_to_old_centroid: list[tuple[Blob, float]] = []

        for blob in blobs_in_frame:
            try:
                indx, centroid, dist = blob.index_and_centroid_closer_to(
                    old_centroid, centroid_id
                )
            except ValueError:  # blob has not centroid_id
                pass
            else:
                dist_to_old_centroid.append((blob, dist))

        try:
            blob_with_old_centroid = min(dist_to_old_centroid, key=lambda x: x[1])[0]
            blob_with_old_centroid.update_centroid(
                old_centroid, new_centroid, centroid_id
            )
        except ValueError:
            logging.error(
                "No blob has the centroid %s in frame %s", old_centroid, frame_number
            )

    def add_centroid(self, frame_number: int, identity: int, centroid):
        centroid = tuple(centroid)
        assert len(centroid) == 2
        blobs_in_frame = self.blobs_in_video[frame_number]
        if not blobs_in_frame:
            self.add_blob(frame_number, centroid, identity)
            return

        for blob in blobs_in_frame:
            if blob.contains_point(centroid):
                blob.add_centroid(centroid, identity)
                return

        blob = min(blobs_in_frame, key=lambda b: b.distance_from_countour_to(centroid))
        blob.add_centroid(centroid, identity)

    def add_blob(self, frame_number: int, centroid: tuple, identity: int):
        """[Validation] Adds a Blob object the frame number.

        Adds a Blob object to a given frame_number with a given centroid and
        identity. Note that this Blob won't have most of the features (e.g.
        area, contour, fragment_identifier, bbox, ...). It is only
        intended to be used for validation and correction of trajectories.
        The new blobs generated are considered to be individuals.

        Parameters
        ----------
        frame_number : int
            Frame in which the new blob will be added
        centroid : tuple
            The centroid of the new blob
        identity : int
            Identity of the new blob
        """
        contour = np.array(
            [
                [centroid[0] - 2, centroid[1] - 2],
                [centroid[0] - 2, centroid[1] + 2],
                [centroid[0] + 2, centroid[1] + 2],
                [centroid[0] + 2, centroid[1] - 2],
            ],
            int,
        )
        new_blob = Blob(contour, frame_number)
        new_blob.added_by_user = True
        new_blob.user_generated_centroids = [(centroid[0], centroid[1])]
        new_blob.user_generated_identities = [identity]
        new_blob.is_an_individual = True
        self.blobs_in_video[frame_number].append(new_blob)

    # Deprecated methods

    @property
    @deprecated(
        version="6.0.0",
        reason="Use `max(map(len, ListOfBlobs.blobs_in_video))` instead",
    )
    def max_number_of_blobs_in_one_frame(self):
        return max(map(len, self.blobs_in_video))

    @deprecated(
        version="6.0.0",
        reason="Loading ListOfBlobs from v4 is not supported anymore",
        action="error",
    )
    def load_from_v4(self): ...

    @deprecated(
        version="6.0.0",
        reason="Use `idtrackerai.base.crossings_detection.crossing_detector._update_id_image_dataset_with_crossings` instead",
        action="error",
    )
    def update_id_image_dataset_with_crossings(self): ...

    @deprecated(
        version="6.0.0",
        reason="Use `idtrackerai.extra_tools.validator.validation_GUI.ValidationGUI.reset_session()` instead",
        action="error",
    )
    def reset_user_generated_corrections(self): ...



import logging
from pathlib import Path
from typing import Literal

import numpy as np

from idtrackerai import (
    Fragment,
    GlobalFragment,
    ListOfFragments,
    ListOfGlobalFragments,
    conf,
)

from ..network import IdentifierBase, get_predictions

AccStrategy = Literal["global", "partial"]


rng = np.random.default_rng()


class AccumulationManager:
    "Manages the process of accumulating images for training the network"

    accumulation_strategy: AccStrategy
    n_animals: int
    list_of_fragments: ListOfFragments
    list_of_global_fragments: ListOfGlobalFragments
    current_step: int
    accumulation_strategy: AccStrategy = "global"
    temporary_used_fragments: set[int]
    accumulation_statistics: dict[str, list[float]]

    used_images: np.ndarray | None = None
    used_labels: np.ndarray | None = None
    new_images: np.ndarray | None = None
    new_labels: np.ndarray | None = None
    ratio_accumulated_images: float = 0

    n_noncertain_global_fragments: int
    n_random_assigned_global_fragments: int
    n_nonconsistent_global_fragments: int
    n_nonunique_global_fragments: int
    n_sparse_fragments: int
    n_noncertain_fragments: int
    n_random_assigned_fragments: int
    n_nonconsistent_fragments: int
    n_nonunique_fragments: int
    n_acceptable_fragments: int

    def __init__(
        self,
        n_animals: int,
        list_of_fragments: ListOfFragments,
        list_of_global_fragments: ListOfGlobalFragments,
    ):
        logging.info("Initializing accumulation manager")

        self.n_animals = n_animals
        self.list_of_fragments = list_of_fragments
        self.list_of_global_fragments = list_of_global_fragments
        self.current_step = 0
        self.reset_accumulation_statistics()

    @property
    def new_global_fragments_for_training(self) -> bool:
        """We stop the accumulation when there are not more global fragments
        that are acceptable for training."""
        there_are = any(
            (
                global_fragment.acceptable_for_training(self.accumulation_strategy)
                and not global_fragment.used_for_training
            )
            for global_fragment in self.list_of_global_fragments
        )

        if there_are:
            log = "There are global fragments acceptable for training"
        else:
            log = "There are no more global fragments acceptable for training"
        if self.accumulation_strategy == "partial":
            log = log.replace("global fragments", "fragments")
        logging.info("[bold]" + log, extra={"markup": True})

        return there_are

    def get_new_images_and_labels(self):
        """Get the images and labels of the new global fragments that are going
        to be used for training. This function checks whether the images of a individual
        fragment have been added before"""

        images = []
        labels = []
        for fragment in self.list_of_fragments.individual_fragments:
            if fragment.acceptable_for_training and not fragment.used_for_training:
                images += fragment.image_locations
                labels += [fragment.temporary_id] * fragment.n_images

        if images:
            self.new_images, self.new_labels = np.asarray(images), np.asarray(labels)
        else:
            self.new_images, self.new_labels = None, None

        n_used_images = len(self.used_images) if self.used_images is not None else 0
        n_new_images = len(self.new_images) if self.new_images is not None else 0
        n_images = n_used_images + n_new_images

        if n_new_images:
            logging.info("%d new images for training", n_new_images)
        else:
            logging.info("There are no new images in this accumulation")

        if n_used_images:
            logging.info("%d old images for training", n_used_images)

        ratio = n_images / self.list_of_fragments.n_images_in_global_fragments
        logging.info(
            f"{n_images} images in total, {ratio:.2%} of the total accumulable"
        )

    def get_old_and_new_images(self):
        "Get a mix of old (already used) and newly labeled images for training the CNN"
        images: list[np.ndarray] = []
        labels: list[int] = []
        for i in range(self.n_animals):
            if self.new_labels is None or self.new_images is None:
                all_new_images = np.empty(0, int)
            else:
                all_new_images = self.new_images[self.new_labels == i]

            if self.used_labels is None or self.used_images is None:
                all_used_images = np.empty(0, int)
            else:
                all_used_images = self.used_images[self.used_labels == i]

            n_new_images = len(all_new_images)
            n_used_images = len(all_used_images)
            n_images_for_individual = n_new_images + n_used_images

            if n_images_for_individual <= conf.MAXIMAL_IMAGES_PER_ANIMAL:
                # if the total number of images for this label does not exceed
                # the conf.MAXIMAL_IMAGES_PER_ANIMAL
                # we take all the new images and all the used images
                if self.new_images is not None:
                    images.append(all_new_images)
                    labels.append(i)
                if self.used_images is not None:
                    # this condition is set because the first time we accumulate
                    # the variable used_images is None
                    images.append(all_used_images)
                    labels.append(i)
                continue

            # we take a proportion of the old images a new images only if the
            # total number of images for this label is bigger than the
            # limit conf.MAXIMAL_IMAGES_PER_ANIMAL
            number_samples_new = int(conf.MAXIMAL_IMAGES_PER_ANIMAL * conf.RATIO_NEW)
            number_samples_used = conf.MAXIMAL_IMAGES_PER_ANIMAL - number_samples_new
            if n_used_images < number_samples_used:
                # if the proportion of used images is bigger than the number of
                # used images we take all the used images for this label and update
                # the number of new images to reach the conf.MAXIMAL_IMAGES_PER_ANIMAL
                number_samples_used = n_used_images
                number_samples_new = (
                    conf.MAXIMAL_IMAGES_PER_ANIMAL - number_samples_used
                )
            if n_new_images < number_samples_new:
                # if the proportion of new images is bigger than the number of
                # new images we take all the new images for this label and update
                # the number of used images to reac the conf.MAXIMAL_IMAGES_PER_ANIMAL
                number_samples_new = n_new_images
                number_samples_used = (
                    conf.MAXIMAL_IMAGES_PER_ANIMAL - number_samples_new
                )
            # we put together a random sample of the new images and the used images
            if self.new_images is not None:
                images.append(
                    rng.choice(all_new_images, number_samples_new, replace=False)
                )
                labels.append(i)
            if self.used_images is not None:
                # this condition is set because the first time we accumulate
                # the variable used_images is None
                images.append(
                    rng.choice(all_used_images, number_samples_used, replace=False)
                )
                labels.append(i)

        labels_array = np.repeat(labels, np.fromiter(map(len, images), dtype=int))
        logging.info(
            "Gathered %d labeled images (new and used) for continuous training",
            len(labels_array),
        )
        return np.concatenate(images), labels_array

    def get_old_images(self):
        """Gather a sample of already used images"""
        if self.used_labels is None or self.used_images is None:
            raise RuntimeError("No used images")

        images: list[np.ndarray] = []
        labels: list[int] = []

        for i in range(self.n_animals):
            all_used_images = self.used_images[self.used_labels == i]
            images.append(
                rng.choice(
                    all_used_images, conf.MAXIMAL_IMAGES_PER_ANIMAL, replace=False
                )
                if len(all_used_images) > conf.MAXIMAL_IMAGES_PER_ANIMAL
                else all_used_images
            )
            labels.append(i)
        labels_array = np.repeat(labels, np.fromiter(map(len, images), dtype=int))
        return np.concatenate(images), labels_array

    def get_new_images(self):
        """Gather a random selection of newly labeled images to train"""
        new_images: list[np.ndarray] = []
        new_labels: list[int] = []
        n_old_images_used = 0
        if self.new_labels is None or self.new_images is None:
            raise RuntimeError("No new imagesC")
        for i in range(self.n_animals):
            new_labels.append(i)
            all_new_images = self.new_images[self.new_labels == i]
            if len(all_new_images) < 4:
                new_images.append(all_new_images)
                if self.used_images is not None:
                    n_old_images_used += 4
                    new_labels.append(i)
                    new_images.append(
                        rng.choice(
                            self.used_images[self.used_labels == i], 4, replace=False
                        )
                    )
            elif len(all_new_images) > conf.MAXIMAL_IMAGES_PER_ANIMAL:
                new_images.append(
                    rng.choice(
                        all_new_images, conf.MAXIMAL_IMAGES_PER_ANIMAL, replace=False
                    )
                )
            else:
                new_images.append(all_new_images)

        new_labels_array = np.repeat(
            new_labels, np.fromiter(map(len, new_images), dtype=int)
        )
        new_images_array = np.concatenate(new_images)
        if n_old_images_used:
            logging.info(
                "%d old images added to balance the training dataset", n_old_images_used
            )
        logging.info(
            "Gathered %d newly labeled images for training", len(new_images_array)
        )
        return new_images_array, new_labels_array

    def update_used_images_and_labels(self):
        """Sets as used the images already used for training"""
        logging.info("Update images and labels used for training")
        if self.current_step == 0:
            assert self.used_images is None
            self.used_images = self.new_images
            self.used_labels = self.new_labels
        elif self.new_images is not None:
            assert self.used_images is not None
            assert self.used_labels is not None
            assert self.new_labels is not None
            self.used_images = np.concatenate((self.used_images, self.new_images))
            self.used_labels = np.concatenate((self.used_labels, self.new_labels))

    def update_fragments_used_for_training(self):
        """Once a global fragment has been used for training, sets the flags
        used_for_training to TRUE and acceptable_for_training to FALSE"""
        logging.info("Updating fragments used for training")
        for gf in self.list_of_global_fragments:
            if gf.acceptable_for_training("global") and not gf.used_for_training:
                gf.accumulation_step = self.current_step

        for fragment in self.list_of_fragments:
            if fragment.acceptable_for_training and not fragment.used_for_training:
                fragment.used_for_training = True
                fragment.acceptable_for_training = False
                fragment.set_partially_or_globally_accumulated(
                    self.accumulation_strategy
                )
                fragment.accumulation_step = self.current_step

    def assign_identities_to_fragments_used_for_training(self):
        """Assign the identities to the global fragments used for training and
        their individual fragments.
        This function checks that the identities of the individual fragments in
        the global fragment
        are consistent with the previously assigned identities
        """
        logging.info("Assigning identities to accumulated global fragments")
        for fragment in self.list_of_fragments:
            if fragment.used_for_training:
                assert fragment.temporary_id is not None
                fragment.identity = fragment.temporary_id + 1
                fragment.P1_vector[:] = 0.0
                fragment.P1_vector[fragment.temporary_id] = 1.0

    def split_predictions_after_network_assignment(
        self,
        predictions: np.ndarray,
        probabilities: np.ndarray,
        indices_to_split: np.ndarray,
        candidate_fragments_identifiers: list[int],
    ):
        """Gathers predictions relative to fragment images from the GPU and
        splits them according to their organization in fragments.
        """
        logging.debug("Computing fragment prediction statistics")
        fragments_predictions = np.split(predictions, indices_to_split)
        fragments_probabilities = np.split(probabilities, indices_to_split)

        for predictions, probabilities_, identifier in zip(
            fragments_predictions,
            fragments_probabilities,
            candidate_fragments_identifiers,
        ):
            self.list_of_fragments.fragments[identifier].set_identification_statistics(
                predictions, probabilities_, self.n_animals
            )

    def reset_accumulation_statistics(self):
        self.accumulation_statistics = {
            "n_accumulated_global_fragments": [],
            "n_non_certain_global_fragments": [],
            "n_randomly_assigned_global_fragments": [],
            "n_nonconsistent_global_fragments": [],
            "n_nonunique_global_fragments": [],
            "n_acceptable_global_fragments": [],
            "ratio_of_accumulated_images": [],
        }

    def update_accumulation_statistics(self):
        stats = self.accumulation_statistics
        stats["n_accumulated_global_fragments"].append(
            sum(
                global_fragment.used_for_training
                for global_fragment in self.list_of_global_fragments
            )
        )
        stats["n_non_certain_global_fragments"].append(
            self.n_noncertain_global_fragments
        )
        stats["n_randomly_assigned_global_fragments"].append(
            self.n_random_assigned_global_fragments
        )
        stats["n_nonconsistent_global_fragments"].append(
            self.n_nonconsistent_global_fragments
        )
        stats["n_nonunique_global_fragments"].append(self.n_nonunique_global_fragments)
        stats["n_acceptable_global_fragments"].append(
            sum(
                global_fragment.acceptable_for_training(self.accumulation_strategy)
                for global_fragment in self.list_of_global_fragments
            )
        )
        stats["ratio_of_accumulated_images"].append(self.ratio_accumulated_images)

    def reset_accumulation_variables(self):
        """After an accumulation is finished reinitialise the variables involved
        in the process.
        """
        self.temporary_used_fragments = set()
        self.n_noncertain_global_fragments = 0
        self.n_random_assigned_global_fragments = 0
        self.n_nonconsistent_global_fragments = 0
        self.n_nonunique_global_fragments = 0
        self.n_sparse_fragments = 0
        self.n_noncertain_fragments = 0
        self.n_random_assigned_fragments = 0
        self.n_nonconsistent_fragments = 0
        self.n_nonunique_fragments = 0
        self.n_acceptable_fragments = 0

    def log_global_accumulation_variables(self):
        lines = (
            "Global prediction results:",
            f"Acceptable global fragments: {self.n_acceptable_global_fragments}",
            "Non acceptable global fragments:"
            f" {self.n_noncertain_global_fragments + self.n_random_assigned_global_fragments + self.n_nonconsistent_global_fragments + self.n_nonunique_global_fragments}",
            f"    Non certain: {self.n_noncertain_global_fragments}",
            f"    Non significant: {self.n_random_assigned_global_fragments}",
            f"    Non consistent: {self.n_nonconsistent_global_fragments}",
            f"    Non unique: {self.n_nonunique_global_fragments}",
        )
        logging.info("\n    ".join(lines))

    def log_partial_accumulation_variables(self):
        lines = (
            "Partial prediction results:",
            f"Acceptable fragments: {self.n_acceptable_fragments}",
            "Non acceptable fragments:"
            f" {self.n_noncertain_fragments + self.n_random_assigned_fragments + self.n_nonconsistent_fragments + self.n_nonunique_fragments + self.n_sparse_fragments}",
            f"    Non certain: {self.n_noncertain_fragments}",
            f"    Non significant: {self.n_random_assigned_fragments}",
            f"    Non consistent: {self.n_nonconsistent_fragments}",
            f"    Non unique: {self.n_nonunique_fragments}",
            f"    Too sparse: {self.n_sparse_fragments}",
        )
        logging.info("\n    ".join(lines))

    def assign_identities(self) -> None:
        """Assigns identities during test to individual fragments and rank them
        according to the score computed from the certainty of identification
        and the minimum distance traveled."""
        self.reset_accumulation_variables()
        self.accumulation_strategy = "global"
        logging.debug(
            "Accumulating by [bold]global[/] strategy", extra={"markup": True}
        )
        for global_fragment in self.list_of_global_fragments:
            self.check_if_is_globally_acceptable_for_training(global_fragment)

        self.n_acceptable_global_fragments = sum(
            global_fragment.acceptable_for_training(self.accumulation_strategy)
            and not global_fragment.used_for_training
            for global_fragment in self.list_of_global_fragments
        )

        self.log_global_accumulation_variables()

        if self.n_acceptable_global_fragments > 0:
            logging.info("Global accumulation succeeded")
            return
        logging.info("Global accumulation failed")

        if (
            self.ratio_accumulated_images
            < conf.MIN_RATIO_OF_IMGS_ACCUMULATED_GLOBALLY_TO_START_PARTIAL_ACCUMULATION
        ):
            logging.info(
                f"The ratio of accumulated images ({self.ratio_accumulated_images:.2%})"
                " is too small and a partial accumulation might fail."
            )
            return

        if self.ratio_accumulated_images > conf.THRESHOLD_EARLY_STOP_ACCUMULATION:
            logging.info(
                f"The ratio of accumulated images ({self.ratio_accumulated_images:.2%})"
                f" is higher than {conf.THRESHOLD_EARLY_STOP_ACCUMULATION:.2%}, early"
                " stopping"
            )
            return

        logging.debug(
            "Accumulating by [bold]partial[/] strategy", extra={"markup": True}
        )
        self.accumulation_strategy = "partial"
        for global_fragment in self.list_of_global_fragments:
            if not global_fragment.used_for_training:
                self.check_if_is_partially_acceptable_for_training(global_fragment)
        self.log_partial_accumulation_variables()

    def reset_non_acceptable_fragment(self, fragment: Fragment):
        """Resets the collection of non-acceptable fragments.

        Parameters
        ----------
        fragment : Fragment object
            Collection of images related to the same individual
        """
        if (
            fragment.identifier not in self.temporary_used_fragments
            and not fragment.used_for_training
        ):
            fragment.temporary_id = None
            fragment.acceptable_for_training = False

    def reset_non_acceptable_global_fragment(self, global_fragment: GlobalFragment):
        """Reset the flag for non-accpetable global fragments.

        Parameters
        ----------
        global_fragment : GlobalFragment object
            Collection of images relative to a part of the video in which all the animals are visible.
        """
        for fragment in global_fragment:
            self.reset_non_acceptable_fragment(fragment)

    def check_if_is_globally_acceptable_for_training(
        self, global_fragment: GlobalFragment
    ):
        if global_fragment.used_for_training:
            return
        assert self.accumulation_strategy == "global"

        for fragment in global_fragment:
            fragment.acceptable_for_training = True

        for fragment in global_fragment:
            if fragment.used_for_training:
                continue
            if not fragment.is_certain:
                # if the certainty of the individual fragment is not high enough
                # we set the global fragment to be non-acceptable for training
                self.reset_non_acceptable_global_fragment(global_fragment)
                self.n_noncertain_global_fragments += 1
                break

        # Compute identities if the global_fragment is certain
        if not global_fragment.acceptable_for_training("global"):
            return

        P1_array, indices_sorted_by_P1 = get_P1_array_and_argsort(global_fragment)
        # set to zero the P1 of the the identities of the individual
        # fragments that have been already used
        for fragment_index, fragment in enumerate(global_fragment):
            if (
                fragment.used_for_training
                or fragment.identifier in self.temporary_used_fragments
            ):
                P1_array[fragment_index] = 0
                P1_array[:, fragment.temporary_id] = 0
            else:  # if fragment's row has already been set to 0, there's no need to filter by incompatible ROIs
                ids_not_compatible_with_roi = (
                    self.list_of_fragments.id_to_exclusive_roi != fragment.exclusive_roi
                )
                P1_array[fragment_index, ids_not_compatible_with_roi] = 0

        # assign temporal identity to individual fragments by hierarchical P1
        for fragment_index in indices_sorted_by_P1:
            fragment: Fragment = global_fragment.fragments[fragment_index]
            if fragment.temporary_id is not None:
                continue
            if p1_below_random(P1_array, fragment_index, fragment):
                fragment.P1_below_random = True
                self.n_random_assigned_global_fragments += 1
                self.reset_non_acceptable_global_fragment(global_fragment)
                break

            temporary_id = P1_array[fragment_index].argmax()
            if fragment.is_inconsistent_with_coexistent_fragments(temporary_id):
                self.reset_non_acceptable_global_fragment(global_fragment)
                fragment.non_consistent = True
                self.n_nonconsistent_global_fragments += 1
                break

            fragment.temporary_id = int(temporary_id)
            P1_array[fragment_index] = 0.0
            P1_array[:, temporary_id] = 0.0

        # Check if the global fragment is unique after assigning the identities
        if not global_fragment.acceptable_for_training("global"):
            return

        if global_fragment.is_unique(self.n_animals):
            self.temporary_used_fragments.update(
                fragment.identifier
                for fragment in global_fragment
                if not fragment.used_for_training
            )
        else:
            # set acceptable_for_training to False and temporary_id to
            # None for all the individual_fragments
            # that had not been accumulated before (i.e. not in
            # temporary_individual_fragments_used or individual_fragments_used)
            self.reset_non_acceptable_global_fragment(global_fragment)
            self.n_nonunique_global_fragments += 1

    def check_if_is_partially_acceptable_for_training(
        self, global_fragment: GlobalFragment
    ):
        assert self.accumulation_strategy == "partial"
        for fragment in global_fragment:
            fragment.acceptable_for_training = False

        for fragment in global_fragment:
            if fragment.used_for_training:
                continue

            if fragment.has_enough_accumulated_coexisting_fragments:
                if not fragment.is_certain:
                    self.reset_non_acceptable_fragment(fragment)
                    self.n_noncertain_fragments += 1
                else:
                    fragment.acceptable_for_training = True
            else:
                self.reset_non_acceptable_fragment(fragment)
                self.n_sparse_fragments += 1

        # Compute identities if the global_fragment is certain
        # get array of P1 values for the global fragment

        P1_array, indices_sorted_by_P1 = get_P1_array_and_argsort(global_fragment)
        # set to zero the P1 of the the identities of the individual
        # fragments that have been already used
        for fragment_index, fragment in enumerate(global_fragment):
            if (
                fragment.used_for_training
                or fragment.identifier in self.temporary_used_fragments
            ):
                P1_array[fragment_index] = 0.0
                P1_array[:, fragment.temporary_id] = 0.0
            else:  # if fragment's row has already been set to 0, there's no need to filter by incompatible ROIs
                ids_not_compatible_with_roi = (
                    self.list_of_fragments.id_to_exclusive_roi != fragment.exclusive_roi
                )
                P1_array[fragment_index, ids_not_compatible_with_roi] = 0

        # assign temporary identity to individual fragments by hierarchical P1
        for fragment_index in indices_sorted_by_P1:
            fragment: Fragment = global_fragment.fragments[fragment_index]

            if (
                fragment.temporary_id is not None
                or not fragment.acceptable_for_training
            ):
                continue

            if p1_below_random(P1_array, fragment_index, fragment):
                self.reset_non_acceptable_fragment(fragment)
                fragment.P1_below_random = True
                self.n_random_assigned_fragments += 1
                continue

            temporary_id = P1_array[fragment_index].argmax()
            if fragment.is_inconsistent_with_coexistent_fragments(temporary_id):
                self.reset_non_acceptable_fragment(fragment)
                fragment.non_consistent = True
                self.n_nonconsistent_fragments += 1
                continue

            fragment.acceptable_for_training = True
            fragment.temporary_id = int(temporary_id)
            P1_array[fragment_index] = 0.0
            P1_array[:, temporary_id] = 0.0

        # Check if the global fragment is unique after assigning the identities
        if not global_fragment.is_partially_unique:
            for fragment in global_fragment:
                if fragment.temporary_id in global_fragment.duplicated_identities:
                    self.reset_non_acceptable_fragment(fragment)
                    self.n_nonunique_fragments += 1

        self.temporary_used_fragments.update(
            fragment.identifier
            for fragment in global_fragment
            if fragment.acceptable_for_training and not fragment.used_for_training
        )
        self.n_acceptable_fragments += sum(
            bool(fragment.acceptable_for_training) and not fragment.used_for_training
            for fragment in global_fragment
        )

        assert all(
            fragment.temporary_id is not None
            for fragment in global_fragment
            if fragment.acceptable_for_training and fragment.is_an_individual
        )


def get_predictions_of_candidates_fragments(
    identification_model: IdentifierBase,
    id_images_file_paths: list[Path],
    list_of_fragments: ListOfFragments,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, list[int]]:
    """Get predictions of individual fragments that have been used to train the
    idCNN in an accumulation's iteration

    Parameters
    ----------
    net : ConvNetwork object
        network used to identify the animals
    session : <Session object>
        Object containing all the parameters of the video.
    fragments : list
        List of fragment objects

    Returns
    -------
    assigner._predictions  : nd.array
        predictions associated to each image organised by individual fragments
    assigner._softmax_probs : np.array
        softmax vector associated to each image organised by individual fragments
    np.cumsum(lengths)[:-1]  : nd.array
        cumulative sum of the number of images contained in every fragment
        (used to rebuild the collection of images per fragment after gathering
        predicions and softmax vectors from the gpu)
    candidate_individual_fragments_identifiers : list
        list of fragment identifiers
    """
    image_locations: list[tuple[int, int]] = []
    lengths: list[int] = []
    candidate_fragments_identifiers: list[int] = []

    for fragment in list_of_fragments.individual_fragments:
        if not fragment.used_for_training:
            image_locations += fragment.image_locations
            lengths.append(fragment.n_images)
            candidate_fragments_identifiers.append(fragment.identifier)

    assert image_locations

    predictions, probabilities = get_predictions(
        identification_model, image_locations, id_images_file_paths
    )

    assert sum(lengths) == len(predictions)
    return (
        predictions,
        probabilities,
        np.cumsum(lengths)[:-1],
        candidate_fragments_identifiers,
    )


def get_P1_array_and_argsort(
    global_fragment: GlobalFragment,
) -> tuple[np.ndarray, np.ndarray]:
    """Given a global fragment computes P1 for each of its individual
    fragments and returns a
    matrix of sorted indices according to P1

    Parameters
    ----------
    global_fragment : GlobalFragment object
        Collection of images relative to a part of the video in which all
        the animals are visible.

    Returns
    -------
    P1_array : nd.array
        P1 computed for every individual fragment in the global fragment
    index_individual_fragments_sorted_by_P1 : nd.array
        Argsort of P1 array of each individual fragment
    """
    P1_array = np.array([fragment.P1_vector for fragment in global_fragment])
    return P1_array, np.argsort(P1_array.max(1))[::-1]


def p1_below_random(
    P1_array: np.ndarray, index_individual_fragment: np.ndarray, fragment: Fragment
):
    """Evaluate if a fragment has been assigned with a certainty lower than
    random (wrt the number of possible identities)

    Parameters
    ----------
    P1_array  : nd.array
        P1 vector of a fragment object
    index_individual_fragment  : nd.array
        Argsort of the P1 array of fragment
    fragment : Fragment
        Fragment object containing images associated with a single individual

    Returns
    -------
    p1_below_random_flag : bool
        True if a fragment has been identified with a certainty below random
    """
    return P1_array[index_individual_fragment].max() < (1.0 / fragment.n_images)



# Each Qt binding is different, so...
# pyright: reportIncompatibleMethodOverride=false
from pathlib import Path
from typing import Literal

import numpy as np
from qtpy.QtCore import Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtCore import QEvent, QPointF, QSettings, Qt
from qtpy.QtGui import (
    QIcon,
    QKeyEvent,
    QPainter,
    QPainterPath,
    QPalette,
    QPolygonF,
    QResizeEvent,
)
from qtpy.QtWidgets import (
    QApplication,
    QDialog,
    QFileDialog,
    QFrame,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QSizePolicy,
    QStyle,
    QToolButton,
    QWidget,
)

from idtrackerai import Session
from idtrackerai.utils import IdtrackeraiError, get_vertices_from_label

ADD_ICON = QIcon.fromTheme("list-add")
REMOVE_ICON = QIcon.fromTheme("edit-clear")

_theme_icons = {
    "cancel": QIcon.fromTheme("dialog-cancel"),
    "ok": QIcon.fromTheme("dialog-ok"),
    "add": QIcon.fromTheme("list-add"),
    "remove": QIcon.fromTheme("edit-clear"),
    "run": QIcon.fromTheme("system-run"),
    "refresh": QIcon.fromTheme("view-refresh"),
    "undo": QIcon.fromTheme("edit-undo"),
}
_style_icons = {
    "cancel": QStyle.StandardPixmap.SP_DialogCancelButton,
    "ok": QStyle.StandardPixmap.SP_DialogOkButton,
    "add": QStyle.StandardPixmap.SP_FileDialogNewFolder,
    "remove": QStyle.StandardPixmap.SP_TrashIcon,
    "run": QStyle.StandardPixmap.SP_ArrowRight,
    "refresh": QStyle.StandardPixmap.SP_BrowserReload,
    "undo": QStyle.StandardPixmap.SP_ArrowBack,
}


def get_icon(
    icon_name: Literal["cancel", "ok", "add", "remove", "run", "refresh", "undo"],
) -> QIcon:
    theme_icon = _theme_icons[icon_name]
    style = QApplication.style()
    if theme_icon.isNull() and style is not None:
        return style.standardIcon(_style_icons[icon_name])
    return theme_icon


class AddBtn(QToolButton):
    """A button to add items"""

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        if ADD_ICON.isNull():
            self.setText("Add")
        else:
            self.setIcon(ADD_ICON)
        self.setToolTip("Add")


class RemoveBtn(QToolButton):
    """A button to remove items"""

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        if REMOVE_ICON.isNull():
            self.setText("Remove")
        else:
            self.setIcon(REMOVE_ICON)
        self.setToolTip("Remove")


class LightPopUp(QDialog):
    """A light version of QMessageBox.warning()"""

    def __init__(self) -> None:
        super().__init__()
        self.setWindowFlags(Qt.WindowType.Popup)
        self.setLayout(QHBoxLayout())
        self.text = QLabel("")
        self.text.setWordWrap(True)
        self.icon = QLabel()

        self.icon.setFixedSize(100, 100)
        self.setFixedWidth(500)

        self.layout().addWidget(self.icon)
        self.layout().addWidget(self.text)

    def warning(self, title: str, text) -> None:
        self.icon.setPixmap(QIcon.fromTheme("dialog-warning").pixmap(70, 70))
        self.text.setText(f"<strong><center>{title}</strong></center><br><br>{text}")
        self.exec()

    def info(self, title: str, text) -> None:
        self.icon.setPixmap(QIcon.fromTheme("dialog-information").pixmap(70, 70))
        self.text.setText(f"<strong><center>{title}</strong></center><br><br>{text}")
        self.exec()

    def keyPressEvent(self, *args, **kwargs) -> None:
        self.close()


def open_session(
    app: QWidget, settings: QSettings, session_path: str | Path | None
) -> Session | None:

    if not session_path or isinstance(session_path, bool):
        file_dialog = QFileDialog()
        settings_key = f"{app.__class__.__name__}_filedialog_state"
        state = settings.value(settings_key, b"")
        if state:
            file_dialog.restoreState(state)
        session_path = file_dialog.getExistingDirectory(
            app, "Open session directory", options=QFileDialog.Option.ShowDirsOnly
        )
        settings.setValue(settings_key, file_dialog.saveState())
        if not session_path:
            return

    user_folder = None
    while True:
        try:
            session = Session.load(
                session_path, user_folder, allow_not_found_video_files=False
            )
        except (FileNotFoundError, ValueError) as exc:
            # Provably the session JSON file was not found or was not JSON readable
            QMessageBox.warning(app, "Loading session error", str(exc))
            return
        except IdtrackeraiError as exc:
            # The video files were not found
            response = QMessageBox.warning(
                app,
                "Video files not found",
                f"{exc}\n\nPlease, open the directory containing the video paths to proceed.",
                QMessageBox.StandardButton.Cancel | QMessageBox.StandardButton.Open,
            )
            if response != QMessageBox.StandardButton.Open:
                return

            file_dialog = QFileDialog()
            settings_key = f"{app.__class__.__name__}_filedialog_state"
            state = settings.value(settings_key, b"")
            if state:
                file_dialog.restoreState(state)
            user_folder = file_dialog.getExistingDirectory(
                app,
                "Select the folder containing the video files",
                options=QFileDialog.Option.ShowDirsOnly,
            )
            settings.setValue(settings_key, file_dialog.saveState())
        else:
            return session


class QHLine(QFrame):
    def __init__(self) -> None:
        super().__init__()
        self.setFrameShape(QFrame.Shape.HLine)
        self.setContentsMargins(10, 0, 10, 0)
        self.setEnabled(False)

    def changeEvent(self, event: QEvent) -> None:
        if event.type() == QEvent.Type.EnabledChange:
            self.setEnabled(False)


def build_ROI_patches_from_list(
    list_of_ROIs: list[str] | str | None, width: int, height: int
) -> QPainterPath:
    path = QPainterPath()

    if list_of_ROIs is None:
        return path

    if isinstance(list_of_ROIs, str):
        list_of_ROIs = [list_of_ROIs]

    path.addRect(-0.5, -0.5, width, height)

    for line in list_of_ROIs:
        path_i = get_path_from_points(get_vertices_from_label(line))

        if line[0] == "+":
            path -= path_i
        elif line[0] == "-":
            path += path_i
        else:
            raise TypeError
    return path


def get_path_from_points(points: np.ndarray) -> QPainterPath:
    path = QPainterPath()
    if points.ndim == 2:
        # some polygons are made from a single point, 1 dimension
        path.addPolygon(QPolygonF([QPointF(x, y) for x, y in points]))
    return path.simplified()


class WrappedLabel(QLabel):
    clicked = Signal()

    def mouseReleaseEvent(self, event):
        if (
            event.button() == Qt.MouseButton.LeftButton
            and not (event.modifiers() & Qt.KeyboardModifier.ShiftModifier)
            and not self.hasSelectedText()
        ):
            self.clicked.emit()
        super().mouseReleaseEvent(event)

    def __init__(
        self,
        text: str = "",
        framed: bool = False,
        align: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignLeft,
    ) -> None:
        super().__init__(text)
        if framed:
            self.setBackgroundRole(QPalette.ColorRole.Base)
            self.setAutoFillBackground(True)
            self.setContentsMargins(5, 3, 5, 3)
        self.setAlignment(align)
        self.setWordWrap(True)
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)

    def set_size(self) -> None:
        self.setMinimumHeight(0)
        self.setMinimumHeight(max(self.heightForWidth(self.width()), 1))

    def resizeEvent(self, a0: QResizeEvent) -> None:
        self.set_size()
        super().resizeEvent(a0)

    def setText(self, text: str) -> None:
        # Add Zero-width space in backslashes for proper word wrapping
        super().setText(text.replace("\\", "\\\u200b"))
        self.set_size()

    def text(self):
        return super().text().replace("\u200b", "")


_ignored_keys = {
    Qt.Key.Key_D,
    Qt.Key.Key_A,
    Qt.Key.Key_Left,
    Qt.Key.Key_Right,
    Qt.Key.Key_0,
    Qt.Key.Key_1,
    Qt.Key.Key_2,
    Qt.Key.Key_3,
    Qt.Key.Key_4,
    Qt.Key.Key_5,
    Qt.Key.Key_6,
    Qt.Key.Key_7,
    Qt.Key.Key_8,
    Qt.Key.Key_9,
}


def key_event_modifier(event: QKeyEvent) -> QKeyEvent | None:
    event.key().numerator
    if event.key() == Qt.Key.Key_W:
        return QKeyEvent(event.type(), Qt.Key.Key_Up, event.modifiers())
    if event.key() == Qt.Key.Key_S:
        return QKeyEvent(event.type(), Qt.Key.Key_Down, event.modifiers())
    if event.key() in _ignored_keys:
        # These keys would be accepted by QTableWidget
        # but we want them to control the VideoPlayer
        event.ignore()
        return None
    return event


class TransparentDisabledOverlay(QWidget):
    """Transparent widget to add to another one which will display self.text when the other is disabled"""

    def __init__(self, text: str, parent: QWidget) -> None:
        super().__init__(parent)
        self.text = text
        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents, True)
        self.setAttribute(Qt.WidgetAttribute.WA_NoSystemBackground, True)
        self.parent_widget = parent
        parent.installEventFilter(self)

    def eventFilter(self, widget, event: QEvent) -> bool:
        if event.type() == QEvent.Type.Resize:
            self.setGeometry(self.parent_widget.rect())
        return False

    def paintEvent(self, a0) -> None:
        if self.isEnabled():
            return
        painter = QPainter(self)
        painter.setPen(
            self.palette().color(QPalette.ColorGroup.Active, QPalette.ColorRole.Text)
        )
        painter.drawText(self.rect(), Qt.AlignmentFlag.AlignCenter, self.text)
